#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see AUTHORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars import _CustomIslamicHolidays
from holidays.calendars.gregorian import MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV
from holidays.groups import (
    ChristianHolidays,
    IslamicHolidays,
    InternationalHolidays,
    StaticHolidays,
)
from holidays.observed_holiday_base import ObservedHolidayBase, SUN_TO_NEXT_MON


class Spain(
    ObservedHolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays, StaticHolidays
):
    """
    Holidays checked with official sources for 2010-2023 only.

    References:
     - https://administracion.gob.es/pag_Home/atencionCiudadana/calendarios.html

     Labor Holidays:
     2010: https://www.boe.es/buscar/doc.php?id=BOE-A-2009-18477
     2011: https://www.boe.es/buscar/doc.php?id=BOE-A-2010-15722
     2012: https://www.boe.es/buscar/doc.php?id=BOE-A-2011-16116
     2013: https://www.boe.es/buscar/doc.php?id=BOE-A-2012-13644
     2014: https://www.boe.es/buscar/doc.php?id=BOE-A-2013-12147
     2015: https://www.boe.es/buscar/doc.php?id=BOE-A-2014-10823
     2016: https://www.boe.es/buscar/doc.php?id=BOE-A-2015-11348
     2017: https://www.boe.es/buscar/doc.php?id=BOE-A-2016-9244
     2018: https://www.boe.es/buscar/doc.php?id=BOE-A-2017-11639
     2019: https://www.boe.es/buscar/doc.php?id=BOE-A-2018-14369
     2020: https://www.boe.es/buscar/doc.php?id=BOE-A-2019-14552
     2021: https://www.boe.es/buscar/doc.php?id=BOE-A-2020-13343
     2022: https://www.boe.es/buscar/doc.php?id=BOE-A-2021-17113
     2023: https://www.boe.es/diario_boe/txt.php?id=BOE-A-2022-16755
     2024: https://www.boe.es/buscar/doc.php?id=BOE-A-2023-22014
    """

    country = "ES"
    default_language = "es"
    # Monday following %s.
    observed_label = tr("Lunes siguiente a %s")
    subdivisions = (
        "AN",  # Andalucía
        "AR",  # Aragón
        "AS",  # Asturias
        "CB",  # Cantabria
        "CE",  # Ceuta
        "CL",  # Castilla y León
        "CM",  # Castilla-La Mancha
        "CN",  # Canarias
        "CT",  # Cataluña
        "EX",  # Extremadura
        "GA",  # Galicia
        "IB",  # Islas Baleares
        "MC",  # Murcia
        "MD",  # Madrid
        "ML",  # Melilla
        "NC",  # Navarra
        "PV",  # País Vasco
        "RI",  # La Rioja
        "VC",  # Valenciana
    )
    supported_languages = ("en_US", "es", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self, cls=SpainIslamicHolidays)
        StaticHolidays.__init__(self, cls=SpainStaticHolidays)
        kwargs.setdefault("observed_rule", SUN_TO_NEXT_MON)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        if self._year not in {2012, 2017, 2023}:
            # New Year's Day.
            self._add_new_years_day(tr("Año nuevo"))

        if self._year not in {2013, 2019}:
            # Epiphany.
            self._add_epiphany_day(tr("Epifanía del Señor"))

        # Good Friday.
        self._add_good_friday(tr("Viernes Santo"))

        if self._year not in {2011, 2016, 2022}:
            # Labor Day.
            self._add_labor_day(tr("Fiesta del Trabajo"))

        if self._year not in {2010, 2021}:
            # Assumption Day.
            self._add_assumption_of_mary_day(tr("Asunción de la Virgen"))

        if self._year != 2014:
            # National Day.
            self._add_holiday_oct_12(tr("Fiesta Nacional de España"))

        if self._year not in {2015, 2020}:
            # All Saints' Day.
            self._add_all_saints_day(tr("Todos los Santos"))

            # Constitution Day.
            self._add_holiday_dec_6(tr("Día de la Constitución Española"))

        if self._year not in {2013, 2019, 2024}:
            # Immaculate Conception.
            self._add_immaculate_conception_day(tr("Inmaculada Concepción"))

        if self._year not in {2011, 2016, 2022}:
            # Christmas Day.
            self._add_christmas_day(tr("Natividad del Señor"))

    def _populate_subdiv_an_public_holidays(self):
        if self._year in {2012, 2017, 2023}:
            self._move_holiday(self._add_new_years_day(tr("Año nuevo")))

        if self._year in {2013, 2019}:
            self._move_holiday(self._add_epiphany_day(tr("Epifanía del Señor")))

        # Andalusia Day.
        self._move_holiday(self._add_holiday_feb_28(tr("Día de Andalucia")))

        # Maundy Thursday.
        self._add_holy_thursday(tr("Jueves Santo"))

        if self._year in {2011, 2016, 2022}:
            self._move_holiday(self._add_labor_day(tr("Fiesta del Trabajo")))

        if self._year in {2010, 2021}:
            self._move_holiday(self._add_assumption_of_mary_day(tr("Asunción de la Virgen")))

        if self._year == 2014:
            self._move_holiday(self._add_holiday_oct_12(tr("Fiesta Nacional de España")))

        if self._year in {2015, 2020}:
            self._move_holiday(self._add_all_saints_day(tr("Todos los Santos")))

        if self._year in {2015, 2020}:
            self._move_holiday(self._add_holiday_dec_6(tr("Día de la Constitución Española")))

        if self._year in {2013, 2019, 2024}:
            self._move_holiday(self._add_immaculate_conception_day(tr("Inmaculada Concepción")))

        if self._year in {2011, 2016, 2022}:
            self._move_holiday(self._add_christmas_day(tr("Natividad del Señor")))

    def _populate_subdiv_ar_public_holidays(self):
        if self._year in {2012, 2017, 2023}:
            self._move_holiday(self._add_new_years_day(tr("Año nuevo")))

        if self._year in {2013, 2019}:
            self._move_holiday(self._add_epiphany_day(tr("Epifanía del Señor")))

        self._add_holy_thursday(tr("Jueves Santo"))

        # St. George's Day.
        self._move_holiday(self._add_saint_georges_day(tr("Día de San Jorge")))

        if self._year in {2011, 2016, 2022}:
            self._move_holiday(self._add_labor_day(tr("Fiesta del Trabajo")))

        if self._year in {2010, 2021}:
            self._move_holiday(self._add_assumption_of_mary_day(tr("Asunción de la Virgen")))

        if self._year == 2014:
            self._move_holiday(self._add_holiday_oct_12(tr("Fiesta Nacional de España")))

        if self._year in {2015, 2020}:
            self._move_holiday(self._add_all_saints_day(tr("Todos los Santos")))

        if self._year in {2015, 2020}:
            self._move_holiday(self._add_holiday_dec_6(tr("Día de la Constitución Española")))

        if self._year in {2013, 2019, 2024}:
            self._move_holiday(self._add_immaculate_conception_day(tr("Inmaculada Concepción")))

        if self._year in {2011, 2016, 2022}:
            self._move_holiday(self._add_christmas_day(tr("Natividad del Señor")))

    def _populate_subdiv_as_public_holidays(self):
        if self._year in {2012, 2017, 2023}:
            self._move_holiday(self._add_new_years_day(tr("Año nuevo")))

        if self._year in {2013, 2019}:
            self._move_holiday(self._add_epiphany_day(tr("Epifanía del Señor")))

        self._add_holy_thursday(tr("Jueves Santo"))

        if self._year in {2011, 2016, 2022}:
            self._move_holiday(self._add_labor_day(tr("Fiesta del Trabajo")))

        if self._year in {2010, 2021}:
            self._move_holiday(self._add_assumption_of_mary_day(tr("Asunción de la Virgen")))

        # Asturia Day.
        self._move_holiday(self._add_holiday_sep_8(tr("Día de Asturias")))

        if self._year == 2014:
            self._move_holiday(self._add_holiday_oct_12(tr("Fiesta Nacional de España")))

        if self._year in {2015, 2020}:
            self._move_holiday(self._add_all_saints_day(tr("Todos los Santos")))

        if self._year in {2015, 2020}:
            self._move_holiday(self._add_holiday_dec_6(tr("Día de la Constitución Española")))

        if self._year in {2013, 2019, 2024}:
            self._move_holiday(self._add_immaculate_conception_day(tr("Inmaculada Concepción")))

        if self._year in {2011, 2016, 2022}:
            self._move_holiday(self._add_christmas_day(tr("Natividad del Señor")))

    def _populate_subdiv_cb_public_holidays(self):
        if self._year == 2013:
            self._move_holiday(self._add_epiphany_day(tr("Epifanía del Señor")))

        if self._year != 2018:
            self._add_holy_thursday(tr("Jueves Santo"))

        if self._year in {2013, 2015, 2019, 2020, 2024}:
            # Easter Monday.
            self._add_easter_monday(tr("Lunes de Pascua"))

        if self._year == 2011:
            self._move_holiday(self._add_labor_day(tr("Fiesta del Trabajo")))

        if self._year in {2012, 2013, 2014, 2019, 2024}:
            # St. James' Day.
            self._add_saint_james_day(tr("Santiago Apóstol"))

        if self._year not in {2012, 2015, 2019, 2024}:
            # Cantabria Institutions Day.
            self._add_holiday_jul_28(tr("Día de las Instituciones de Cantabria"))

        if self._year not in {2013, 2019, 2024}:
            # Our Lady of the Bien Aparecida.
            self._add_holiday_sep_15(tr("La Bien Aparecida"))

        if self._year == 2015:
            self._move_holiday(self._add_all_saints_day(tr("Todos los Santos")))

        if self._year == 2019:
            self._move_holiday(self._add_immaculate_conception_day(tr("Inmaculada Concepción")))

        if self._year in {2016, 2022}:
            self._move_holiday(self._add_christmas_day(tr("Natividad del Señor")))

    def _populate_subdiv_ce_public_holidays(self):
        if self._year == 2012:
            self._move_holiday(self._add_new_years_day(tr("Año nuevo")))

        if self._year in {2013, 2019}:
            self._move_holiday(self._add_epiphany_day(tr("Epifanía del Señor")))

        self._add_holy_thursday(tr("Jueves Santo"))

        if self._year == 2011:
            self._move_holiday(self._add_labor_day(tr("Fiesta del Trabajo")))

        if self._year >= 2022:
            # Santa Maria of Africa.
            self._add_holiday_aug_5(tr("Nuestra Señora de África"))

        if self._year not in {2011, 2012, 2015, 2018}:
            # Ceuta Day.
            self._add_holiday_sep_2(tr("Día de Ceuta"))

        if self._year == 2014:
            self._move_holiday(self._add_holiday_oct_12(tr("Fiesta Nacional de España")))

        if self._year == 2015:
            self._move_holiday(self._add_all_saints_day(tr("Todos los Santos")))

        if self._year in {2015, 2020}:
            self._move_holiday(self._add_holiday_dec_6(tr("Día de la Constitución Española")))

        if self._year == 2013:
            self._move_holiday(self._add_immaculate_conception_day(tr("Inmaculada Concepción")))

        if self._year in {2011, 2016}:
            self._move_holiday(self._add_christmas_day(tr("Natividad del Señor")))

        # Eid al-Adha.
        name = tr("Fiesta del Sacrificio-Eidul Adha")
        if self._year == 2011:
            self._add_eid_al_adha_day_two(name)
        elif self._year in {2012, 2014}:
            self._add_eid_al_adha_day_three(name)
        elif self._year >= 2010:
            self._add_eid_al_adha_day(name)

    def _populate_subdiv_cl_public_holidays(self):
        if self._year in {2017, 2023}:
            self._move_holiday(self._add_new_years_day(tr("Año nuevo")))

        if self._year in {2013, 2019}:
            self._move_holiday(self._add_epiphany_day(tr("Epifanía del Señor")))

        if self._year in {2010, 2012}:
            self._add_saint_josephs_day(tr("San José"))

        self._add_holy_thursday(tr("Jueves Santo"))

        if self._year != 2023:
            # Castile and León Day.
            self._move_holiday(self._add_holiday_apr_23(tr("Fiesta de Castilla y León")))

        if self._year in {2016, 2022}:
            self._move_holiday(self._add_labor_day(tr("Fiesta del Trabajo")))

        if self._year in {2011, 2023}:
            self._add_saint_james_day(tr("Santiago Apóstol"))

        if self._year == 2021:
            self._move_holiday(self._add_assumption_of_mary_day(tr("Asunción de la Virgen")))

        if self._year == 2014:
            self._move_holiday(self._add_holiday_oct_12(tr("Fiesta Nacional de España")))

        if self._year in {2015, 2020}:
            self._move_holiday(self._add_all_saints_day(tr("Todos los Santos")))

        if self._year in {2015, 2020}:
            self._move_holiday(self._add_holiday_dec_6(tr("Día de la Constitución Española")))

        if self._year in {2013, 2019, 2024}:
            self._move_holiday(self._add_immaculate_conception_day(tr("Inmaculada Concepción")))

        if self._year in {2011, 2016, 2022}:
            self._move_holiday(self._add_christmas_day(tr("Natividad del Señor")))

    def _populate_subdiv_cm_public_holidays(self):
        if self._year == 2013:
            self._move_holiday(self._add_epiphany_day(tr("Epifanía del Señor")))

        if self._year in {2010, 2011, 2020}:
            # St. Joseph's Day.
            self._add_saint_josephs_day(tr("San José"))

        self._add_holy_thursday(tr("Jueves Santo"))

        if self._year in {2014, 2015, 2019, 2020}:
            self._add_easter_monday(tr("Lunes de Pascua"))

        if self._year not in {2010, 2018}:
            # Corpus Christi.
            self._add_corpus_christi_day(tr("Corpus Christi"))

        if self._year not in {2015, 2020}:
            # Castilla-La Mancha Day.
            self._add_holiday_may_31(tr("Día de Castilla-La Mancha"))

        if self._year == 2015:
            self._move_holiday(self._add_holiday_dec_6(tr("Día de la Constitución Española")))

        if self._year in {2016, 2022}:
            self._move_holiday(self._add_christmas_day(tr("Natividad del Señor")))

    def _populate_subdiv_cn_public_holidays(self):
        if self._year in {2013, 2019}:
            self._move_holiday(self._add_epiphany_day(tr("Epifanía del Señor")))

        self._add_holy_thursday(tr("Jueves Santo"))

        if self._year == 2016:
            self._move_holiday(self._add_labor_day(tr("Fiesta del Trabajo")))

        if self._year != 2021:
            # Day of the Canary Islands.
            self._move_holiday(self._add_holiday_may_30(tr("Día de Canarias")))

        if self._year == 2021:
            self._move_holiday(self._add_assumption_of_mary_day(tr("Asunción de la Virgen")))

        if self._year == 2015:
            self._move_holiday(self._add_all_saints_day(tr("Todos los Santos")))

        if self._year == 2020:
            self._move_holiday(self._add_holiday_dec_6(tr("Día de la Constitución Española")))

        if self._year in {2011, 2022}:
            self._move_holiday(self._add_christmas_day(tr("Natividad del Señor")))

    def _populate_subdiv_ct_public_holidays(self):
        self._add_easter_monday(tr("Lunes de Pascua"))

        if self._year in {2011, 2016, 2022}:
            # Whit Monday.
            self._add_whit_monday(tr("Día de la Pascua Granada"))

        if self._year not in {2012, 2018}:
            # St. John the Baptist.
            self._add_saint_johns_day(tr("San Juan"))

        if self._year not in {2011, 2022}:
            # National Day of Catalonia.
            self._add_holiday_sep_11(tr("Fiesta Nacional de Cataluña"))

        if self._year not in {2010, 2021}:
            # St. Stephen's Day.
            self._add_christmas_day_two(tr("San Esteban"))

    def _populate_subdiv_ex_public_holidays(self):
        if self._year == 2012:
            self._move_holiday(self._add_new_years_day(tr("Año nuevo")))

        if self._year in {2013, 2019}:
            self._move_holiday(self._add_epiphany_day(tr("Epifanía del Señor")))

        if self._year in {2023, 2024}:
            # Shrove Tuesday.
            self._add_carnival_tuesday(tr("Martes de Carnaval"))

        if self._year in {2010, 2017, 2021}:
            # St. Joseph's Day.
            self._move_holiday(self._add_saint_josephs_day(tr("San José")))

        self._add_holy_thursday(tr("Jueves Santo"))

        if self._year in {2011, 2016, 2022}:
            self._move_holiday(self._add_labor_day(tr("Fiesta del Trabajo")))

        if self._year != 2024:
            # Extremadura Day.
            self._move_holiday(self._add_holiday_sep_8(tr("Día de Extremadura")))

        if self._year == 2014:
            self._move_holiday(self._add_holiday_oct_12(tr("Fiesta Nacional de España")))

        if self._year in {2015, 2020}:
            self._move_holiday(self._add_all_saints_day(tr("Todos los Santos")))

        if self._year in {2015, 2020}:
            self._move_holiday(self._add_holiday_dec_6(tr("Día de la Constitución Española")))

        if self._year in {2013, 2019, 2024}:
            self._move_holiday(self._add_immaculate_conception_day(tr("Inmaculada Concepción")))

        if self._year in {2011, 2016, 2022}:
            self._move_holiday(self._add_christmas_day(tr("Natividad del Señor")))

    def _populate_subdiv_ga_public_holidays(self):
        if self._year in {2010, 2011} or 2019 <= self._year <= 2021:
            self._move_holiday(self._add_saint_josephs_day(tr("San José")))

        self._add_holy_thursday(tr("Jueves Santo"))

        if self._year not in {2015, 2020}:
            # Galician Literature Day.
            self._add_holiday_may_17(tr("Día de las Letras Gallegas"))

        if self._year in {2013, 2016, 2020, 2022}:
            self._add_saint_johns_day(tr("San Juan"))

        if self._year != 2021:
            # Galician National Day.
            self._add_holiday_jul_25(tr("Día Nacional de Galicia"))

        if self._year == 2015:
            self._move_holiday(self._add_all_saints_day(tr("Todos los Santos")))

    def _populate_subdiv_ib_public_holidays(self):
        if self._year not in {2015, 2020}:
            # Day of the Balearic Islands.
            self._add_holiday_mar_1(tr("Día de las Islas Baleares"))

        self._add_holy_thursday(tr("Jueves Santo"))

        if self._year != 2014:
            self._add_easter_monday(tr("Lunes de Pascua"))

        if self._year == 2015:
            self._move_holiday(self._add_all_saints_day(tr("Todos los Santos")))

        if self._year in {2015, 2020}:
            self._move_holiday(self._add_holiday_dec_6(tr("Día de la Constitución Española")))

        if self._year in {2011, 2016, 2022}:
            self._move_holiday(self._add_christmas_day(tr("Natividad del Señor")))

        if self._year in {2013, 2014, 2019, 2020}:
            self._add_christmas_day_two(tr("San Esteban"))

    def _populate_subdiv_mc_public_holidays(self):
        if self._year in {2017, 2023}:
            self._move_holiday(self._add_new_years_day(tr("Año nuevo")))

        if self._year in {2013, 2019}:
            self._move_holiday(self._add_epiphany_day(tr("Epifanía del Señor")))

        if (self._year <= 2021 and self._year != 2017) or self._year == 2024:
            self._move_holiday(self._add_saint_josephs_day(tr("San José")))

        self._add_holy_thursday(tr("Jueves Santo"))

        if self._year in {2011, 2022}:
            self._move_holiday(self._add_labor_day(tr("Fiesta del Trabajo")))

        if self._year not in {2013, 2024}:
            # Murcia Day.
            self._move_holiday(self._add_holiday_jun_9(tr("Día de la Región de Murcia")))

        if self._year in {2015, 2020}:
            self._move_holiday(self._add_holiday_dec_6(tr("Día de la Constitución Española")))

        if self._year in {2013, 2024}:
            self._move_holiday(self._add_immaculate_conception_day(tr("Inmaculada Concepción")))

        if self._year in {2016, 2022}:
            self._move_holiday(self._add_christmas_day(tr("Natividad del Señor")))

    def _populate_subdiv_md_public_holidays(self):
        if self._year in {2013, 2019}:
            self._move_holiday(self._add_epiphany_day(tr("Epifanía del Señor")))

        if self._year in {2010, 2012, 2015, 2017, 2021, 2023}:
            self._move_holiday(self._add_saint_josephs_day(tr("San José")))

        self._add_holy_thursday(tr("Jueves Santo"))

        if self._year != 2010:
            # Madrid Day.
            self._move_holiday(self._add_holiday_may_2(tr("Fiesta de la Comunidad de Madrid")))

        if self._year in {2010, 2011, 2014}:
            self._add_corpus_christi_day(tr("Corpus Christi"))

        if self._year in {2011, 2016, 2022, 2024}:
            self._add_saint_james_day(tr("Santiago Apóstol"))

        if self._year == 2020:
            self._move_holiday(self._add_all_saints_day(tr("Todos los Santos")))

        if self._year == 2020:
            self._move_holiday(self._add_holiday_dec_6(tr("Día de la Constitución Española")))

        if self._year == 2019:
            self._move_holiday(self._add_immaculate_conception_day(tr("Inmaculada Concepción")))

        if self._year in {2016, 2022}:
            self._move_holiday(self._add_christmas_day(tr("Natividad del Señor")))

    def _populate_subdiv_ml_public_holidays(self):
        if self._year == 2017:
            self._move_holiday(self._add_new_years_day(tr("Año nuevo")))

        if self._year in {2013, 2019}:
            self._move_holiday(self._add_epiphany_day(tr("Epifanía del Señor")))

        if self._year in {2020, 2021}:
            # Statute of Autonomy of Melilla Day.
            self._add_holiday_mar_13(tr("Estatuto de Autonomía de la Ciudad de Melilla"))

        if self._year <= 2016:
            self._add_saint_josephs_day(tr("San José"))

        self._add_holy_thursday(tr("Jueves Santo"))

        if self._year in {2015, 2020}:
            self._move_holiday(self._add_holiday_dec_6(tr("Día de la Constitución Española")))

        if self._year in {2019, 2024}:
            self._move_holiday(self._add_immaculate_conception_day(tr("Inmaculada Concepción")))

        if self._year in {2011, 2016, 2022}:
            self._move_holiday(self._add_christmas_day(tr("Natividad del Señor")))

        if self._year in {2022, 2023}:
            # Eid al-Fitr.
            self._add_eid_al_fitr_day(tr("Fiesta del Eid Fitr"))

        # Eid al-Adha.
        name = tr("Fiesta del Sacrificio-Aid Al Adha")
        if self._year in {2011, 2012, 2021}:
            self._add_eid_al_adha_day_two(name)
        elif self._year == 2022:
            self._add_eid_al_adha_day_three(name)
        elif self._year >= 2010:
            self._add_eid_al_adha_day(name)

    def _populate_subdiv_nc_public_holidays(self):
        if self._year in {2013, 2019}:
            self._move_holiday(self._add_epiphany_day(tr("Epifanía del Señor")))

        if self._year in {2010, 2012, 2014, 2015, 2019, 2020, 2021}:
            self._add_saint_josephs_day(tr("San José"))

        self._add_holy_thursday(tr("Jueves Santo"))

        self._add_easter_monday(tr("Lunes de Pascua"))

        if self._year in {2011, 2013, 2015, 2016, 2017} or self._year >= 2022:
            self._add_saint_james_day(tr("Santiago Apóstol"))

        if self._year == 2020:
            self._move_holiday(self._add_holiday_dec_6(tr("Día de la Constitución Española")))

        if self._year in {2011, 2016, 2022}:
            self._move_holiday(self._add_christmas_day(tr("Natividad del Señor")))

    def _populate_subdiv_pv_public_holidays(self):
        if self._year in {2010, 2015, 2019, 2020, 2021}:
            self._add_saint_josephs_day(tr("San José"))

        self._add_holy_thursday(tr("Jueves Santo"))

        self._add_easter_monday(tr("Lunes de Pascua"))

        if self._year not in {2010, 2012, 2014, 2018, 2021}:
            self._add_saint_james_day(tr("Santiago Apóstol"))

        if 2011 <= self._year <= 2014:
            # País Vasco Day.
            self._add_holiday_oct_25(tr("Día del País Vasco"))

    def _populate_subdiv_ri_public_holidays(self):
        if self._year in {2010, 2012}:
            self._add_saint_josephs_day(tr("San José"))

        self._add_holy_thursday(tr("Jueves Santo"))

        if self._year not in {2010, 2012, 2018}:
            self._add_easter_monday(tr("Lunes de Pascua"))

        # La Rioja Day.
        self._move_holiday(self._add_holiday_jun_9(tr("Día de La Rioja")))

        if self._year in {2011, 2016}:
            self._add_saint_james_day(tr("Santiago Apóstol"))

        if self._year in {2015, 2020}:
            self._move_holiday(self._add_holiday_dec_6(tr("Día de la Constitución Española")))

        if self._year in {2013, 2019}:
            self._move_holiday(self._add_immaculate_conception_day(tr("Inmaculada Concepción")))

        if self._year == 2022:
            self._move_holiday(self._add_christmas_day(tr("Natividad del Señor")))

    def _populate_subdiv_vc_public_holidays(self):
        if (self._year <= 2022 and self._year != 2017) or self._year == 2024:
            self._add_saint_josephs_day(tr("San José"))

        if self._year in {2011, 2016, 2017, 2022}:
            self._add_holy_thursday(tr("Jueves Santo"))

        self._add_easter_monday(tr("Lunes de Pascua"))

        if self._year == 2011:
            self._move_holiday(self._add_labor_day(tr("Fiesta del Trabajo")))

        if self._year >= 2019:
            self._add_saint_johns_day(tr("San Juan"))

        if self._year not in {2011, 2016, 2022}:
            # Valencian Community Day.
            self._add_holiday_oct_9(tr("Día de la Comunidad Valenciana"))

        if self._year == 2015:
            self._move_holiday(self._add_holiday_dec_6(tr("Día de la Constitución Española")))

        if self._year == 2016:
            self._move_holiday(self._add_christmas_day(tr("Natividad del Señor")))


class ES(Spain):
    pass


class ESP(Spain):
    pass


class SpainIslamicHolidays(_CustomIslamicHolidays):
    EID_AL_ADHA_DATES = {
        2010: (NOV, 17),
        2011: (NOV, 6),
        2012: (OCT, 25),
        2013: (OCT, 15),
        2014: (OCT, 4),
        2015: (SEP, 25),
        2016: (SEP, 12),
        2017: (SEP, 1),
        2018: (AUG, 22),
        2019: (AUG, 12),
        2020: (JUL, 31),
        2021: (JUL, 20),
        2022: (JUL, 9),
        2023: (JUN, 29),
        2024: (JUN, 17),
    }

    EID_AL_FITR_DATES = {
        2022: (MAY, 3),
        2023: (APR, 21),
    }


class SpainStaticHolidays:
    special_ga_public_holidays = {
        # Day following St. Joseph's Day.
        2015: (MAR, 20, tr("Día siguiente a San José")),
    }

    special_md_public_holidays = {
        # St. Joseph's Day Transfer.
        2013: (MAR, 18, tr("Traslado de San José")),
    }

    special_pv_public_holidays = {
        # 80th Anniversary of the first Basque Government.
        2016: (OCT, 7, tr("80 Aniversario del primer Gobierno Vasco")),
        # V Centennial of the Circumnavigation of the World.
        2022: (SEP, 6, tr("V Centenario Vuelta al Mundo")),
    }

    special_vc_public_holidays = {
        # The Fallas.
        2013: (MAR, 18, tr("Lunes de Fallas")),
    }



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see AUTHORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars import _CustomIslamicHolidays
from holidays.calendars.gregorian import JAN, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC
from holidays.constants import BANK, PUBLIC
from holidays.groups import (
    ChristianHolidays,
    InternationalHolidays,
    IslamicHolidays,
    StaticHolidays,
)
from holidays.holiday_base import HolidayBase


class Tanzania(
    HolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays, StaticHolidays
):
    """
    References:
    - https://old.tanzlii.org/tz/legislation/act/1962/48/  # 1962
    - https://old.tanzlii.org/tz/legislation/act/1964/52/  # 1964
    - https://old.tanzlii.org/tz/legislation/act/1966/39/  # 1966
    - https://www.parliament.go.tz/polis/uploads/bills/acts/1566639469-The%20Written%20Laws%20(Miscellaneous%20Amendments)%20Act,%201993.pdf  # 1993  # noqa: E501
    - https://www.parliament.go.tz/polis/uploads/bills/acts/1566638051-The%20Written%20Laws%20(Miscellaneous%20Amendments)%20(No.%202)%20Act,%201994.pdf  # 1994  # noqa: E501
    - http://parliament.go.tz/polis/uploads/bills/acts/1454076376-ActNo-25-2002.pdf  # 2002
    - https://en.wikipedia.org/wiki/Public_holidays_in_Tanzania
    - http://mytanzaniatimes.blogspot.com/2014/08/holiday-nane-nane-farmers-day-in.html
    - https://www.theeastafrican.co.ke/tea/business/tanzania-declares-public-holiday-on-census-day-3918836  # noqa: E501
    - https://www.dw.com/en/samia-suluhu-hassan-who-is-tanzanias-new-president/a-56932264

    Checked With:
    - https://www.bot.go.tz/webdocs/Other/2023%20public%20holidays.pdf  # 2023
    - https://www.bot.go.tz/webdocs/Other/PUBLIC%20HOLIDAYS%202022.pdf  # 2022
    - https://www.bot.go.tz/webdocs/Other/PUBLIC%20HOLIDAYS%202021.pdf  # 2021
    - https://www.bot.go.tz/webdocs/Other/PUBLIC%20HOLIDAYS%202020.pdf  # 2020
    - https://issamichuzi.blogspot.com/2017/11/sikukuu-za-kitaifa-zenye-mapumziko-kwa.html  # 2018
    - https://www.timeanddate.com/holidays/tanzania/ (from 2013 onwards)

    Limitations:

    - Only works from 1994 onwards due to the lack of sources for certain legislation:
        Government Notices No. 79 of 1977
        Government Notices No. 300 of 1985
        Government Notices No. 296 of 1994

    - Exact Islamic holidays dates are only available for 2013-2023; the rest are estimates.
    """

    country = "TZ"
    supported_categories = (BANK, PUBLIC)
    default_language = "sw"
    # %s (estimated).
    estimated_label = tr("%s (makisio)")
    supported_languages = ("en_US", "sw")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self, cls=TanzaniaIslamicHolidays)
        StaticHolidays.__init__(self, TanzaniaStaticHolidays)
        super().__init__(*args, **kwargs)

    def _populate_bank_holidays(self):
        # Written Law (Miscellaneous Amendments) (No. 2) Act, 1994.
        if self._year <= 1993:
            return None

        # Sikukuu ya Pasaka.
        # Status: In-Use.
        # Only observed by financial institutions.

        # Easter Sunday.
        self._add_easter_sunday(tr("Sikukuu ya Pasaka"))

    def _populate_public_holidays(self):
        # Written Law (Miscellaneous Amendments) (No. 2) Act, 1994.
        if self._year <= 1993:
            return None

        # In-lieus ("Badala ya %s") are observed on Monday should it fall on the weekends.
        # Abrogated in Public Holidays Ordinance  No. 28 of 1966.
        # Reinstituted in Written Law (Miscellaneous Amendments) (No. 2) Act, 1994.
        # Claimed to be abrogated again on an unspecified date.

        # Fixed Date Holidays.

        # Mwaka Mpya.
        # Status: In-Use.
        # Abrogated in Public Holidays Ordinance No. 28 of 1966.
        # Reinstituted prior to the Written Law (Miscellaneous Amendments) (No. 1) Act, 1993.

        # New Year's Day.
        self._add_new_years_day(tr("Mwaka Mpya"))

        # Mapinduzi ya Zanzibar.
        # Status: In-Use.
        # Commemorates the fall of the Sultanate of Zanzibar on Jan 12, 1964.

        # Zanzibar Revolution Day.
        self._add_holiday_jan_12(tr("Mapinduzi ya Zanzibar"))

        # Siku ya kumbukumbu ya Rais wa Kwanza wa Serikali ya Mapinduzi Zanzibar
        # Sheikh Abeid Amani Karume.
        # Status: In-Use.
        # Commemorates the death of Abeid Amani Karume on Apr 7, 1972.
        # Assumed to start via Government Notices 79 of 1977.

        self._add_holiday_apr_7(
            # The Sheikh Abeid Amani Karume Day.
            tr(
                "Siku ya kumbukumbu ya Rais wa Kwanza wa Serikali ya Mapinduzi Zanzibar "
                "Sheikh Abeid Amani Karume"
            )
        )

        # Muungano wa Tanganyika na Zanzibar.
        # Status: In-Use.
        # Commemorates the creation of the United Republic of Tanzania on Apr 26, 1964.

        # Union Celebrations.
        self._add_holiday_apr_26(tr("Muungano wa Tanganyika na Zanzibar"))

        # Sikukuu ya Wafanyakazi.
        # Status: In-Use.

        # Worker's Day.
        self._add_labor_day(tr("Sikukuu ya Wafanyakazi"))

        # Sabasaba.
        # Status: In-Use.
        # Celebrates the formation of Tanganyika African National Union party on Jul 7, 1954.
        # First started in the Public Holidays Ordinance No. 57 of 1964.
        # Abrogated in 1993 (need official source on this)
        # Reinstituted via Government Notices 296 of 1994.

        # International Trade Fair.
        self._add_holiday_jul_7(tr("Sabasaba"))

        # Siku ya Wakulima.
        # Status: In-Use.
        # Also known as Nane Nane.
        # Started via Government Notices 296 of 1994 to replace Sabasaba.

        # Peasants Day.
        self._add_holiday_aug_8(tr("Siku ya Wakulima"))

        # Kumbukumbu ya Mwalimu Nyerere.
        # Status: In-Use.
        # Commemorates the death of Julius Kambarage Nyerere on Oct 14, 1999.
        # Adopted in Written Law (Miscellaneous Amendments) (No. 3) Act, 2002.

        if self._year >= 2003:
            # The Mwalimu Nyerere Day.
            self._add_holiday_oct_14(tr("Kumbukumbu ya Mwalimu Nyerere"))

        # Uhuru na Jamhuri.
        # Status: In-Use.
        # Commemorates the Independence of Tanganyika from the United Kingdom on Dec 9, 1961.

        # Independence and Republic Day.
        self._add_holiday_dec_9(tr("Uhuru na Jamhuri"))

        # Kuzaliwa Kristo.
        # Status: In-Use.

        # Christmas Day.
        self._add_christmas_day(tr("Kuzaliwa Kristo"))

        # Siku ya Kupeana Zawadi.
        # Status: In-Use.
        # Abrogated in Public Holidays Ordinance No. 28 of 1966.
        # Reinstituted in Written Law (Miscellaneous Amendments) (No. 1) Act, 1993.

        # Boxing Day.
        self._add_christmas_day_two(tr("Siku ya Kupeana Zawadi"))

        # Christian Calendar Holidays.

        # Ijumaa Kuu.
        # Status: In-Use.

        # Good Friday.
        self._add_good_friday(tr("Ijumaa Kuu"))

        # Jumatatu ya Pasaka.
        # Status: In-Use.

        # Easter Monday.
        self._add_easter_monday(tr("Jumatatu ya Pasaka"))

        # Islamic Calendar Holidays.

        # Eid El-Fitry.
        # Status: In-Use.
        # Used to be celebrated for 2 days in the 1964 and 1966 amendments.

        # Eid al-Fitr.
        self._add_eid_al_fitr_day(tr("Eid El-Fitry"))

        # Eid El-Hajj.
        # Status: In-Use.
        # Used to be celebrated for 2 days in the 1964 amendments.

        # Eid al-Adha.
        self._add_eid_al_adha_day(tr("Eid El-Hajj"))

        # Maulidi.
        # Status: In-Use.

        # Maulid Day.
        self._add_mawlid_day(tr("Maulidi"))

        # Defunct Holidays.

        # 5th Day of February ??? (Name Unavailable)
        # Status: Defunct.
        # Abrogated in Written Law (Miscellaneous Amendments) (No. 1) Act, 1993.


class TZ(Tanzania):
    pass


class TZA(Tanzania):
    pass


class TanzaniaIslamicHolidays(_CustomIslamicHolidays):
    EID_AL_ADHA_DATES = {
        2013: (OCT, 15),
        2014: (OCT, 5),
        2015: (SEP, 23),
        2016: (SEP, 16),
        2017: (SEP, 2),
        2018: (AUG, 22),
        2019: (AUG, 13),
        2020: (JUL, 31),
        2021: (JUL, 21),
        2022: (JUL, 10),
        2023: (JUN, 29),
    }

    EID_AL_FITR_DATES = {
        2013: (AUG, 8),
        2014: (JUL, 29),
        2015: (JUL, 18),
        2016: (JUL, 7),
        2017: (JUN, 26),
        2018: (JUN, 15),
        2019: (JUN, 5),
        2020: (MAY, 24),
        2021: (MAY, 14),
        2022: (MAY, 3),
        2023: (APR, 22),
        2024: (APR, 10),
    }

    MAWLID_DATES = {
        2013: (JAN, 24),
        2014: (JAN, 14),
        2015: ((JAN, 3), (DEC, 24)),
        2016: (DEC, 12),
        2017: (DEC, 1),
        2018: (NOV, 21),
        2019: (NOV, 10),
        2020: (OCT, 29),
        2021: (OCT, 19),
        2022: (OCT, 9),
        2023: (SEP, 28),
    }


class TanzaniaStaticHolidays:
    # Special Holidays.

    # John Pombe Magufuli Inauguration Day.
    john_magufuli_inauguration = tr("Sikukuu ya Kuapishwa kwa John Pombe Magufuli")

    # Tanzania General Election Day.
    tz_election_day = tr("Sikukuu ya Uchaguzi Mkuu wa Tanzania")

    # National Population and Housing Census Day.
    phc_census_day = tr("Siku ya Sensa ya Kitaifa ya Watu na Makazi")

    # John Pombe Magufuli's Funeral.
    john_magufuli_funeral = tr("Mazishi cha John Pombe Magufuli")

    special_public_holidays = {
        2002: (AUG, 25, phc_census_day),
        2015: (NOV, 5, john_magufuli_inauguration),
        2020: (OCT, 28, tz_election_day),
        2021: (
            (MAR, 22, john_magufuli_funeral),
            (MAR, 25, john_magufuli_funeral),
        ),
        2022: (AUG, 23, phc_census_day),
    }



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see AUTHORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr
from typing import Set

from holidays.calendars.gregorian import JAN, FEB, SEP, NOV, THU, FRI, SAT
from holidays.groups import IslamicHolidays, StaticHolidays
from holidays.observed_holiday_base import (
    ObservedHolidayBase,
    THU_TO_PREV_WED,
    FRI_TO_PREV_THU,
    FRI_TO_NEXT_SAT,
    SAT_TO_NEXT_SUN,
    THU_FRI_TO_NEXT_WORKDAY,
    FRI_SAT_TO_NEXT_WORKDAY,
)


class SaudiArabia(ObservedHolidayBase, IslamicHolidays, StaticHolidays):
    """
    There are only 4 official national holidays in Saudi:
        - https://laboreducation.hrsd.gov.sa/en/gallery/274
        - https://laboreducation.hrsd.gov.sa/en/labor-education/322
        - https://english.alarabiya.net/News/gulf/2022/01/27/Saudi-Arabia-to-commemorate-Founding-Day-on-Feb-22-annually-Royal-order  # noqa: E501

    The national day and the founding day holidays are based on the
    Georgian calendar while the other two holidays are based on the
    Islamic Calendar, and they are estimates as they announced each
    year and based on moon sightings; they are:
    - Eid al-Fitr
    - Eid al-Adha

    """

    country = "SA"
    default_language = "ar"
    # %s (estimated).
    estimated_label = tr("(تقدير) %s")
    # %s (observed).
    observed_label = tr("(ملاحظة) %s")
    # %s (observed, estimated).
    observed_estimated_label = tr("(تقدير ملاحظة) %s")
    supported_languages = ("ar", "en_US")

    def __init__(self, *args, **kwargs):
        IslamicHolidays.__init__(self)
        StaticHolidays.__init__(self, SaudiArabiaStaticHolidays)
        kwargs.setdefault("observed_rule", FRI_TO_PREV_THU + SAT_TO_NEXT_SUN)
        super().__init__(*args, **kwargs)

    def _add_islamic_observed(self, dts: Set[date]) -> None:
        # Observed days are added to make up for any days falling on a weekend.
        if not self.observed:
            return None
        observed_rule = THU_FRI_TO_NEXT_WORKDAY if self._year <= 2012 else FRI_SAT_TO_NEXT_WORKDAY
        for dt in dts:
            for i in range(4):
                self._add_observed(dt + td(days=-i), name=self[dt], rule=observed_rule)

    def _populate_public_holidays(self):
        # Weekend used to be THU, FRI before June 28th, 2013.
        # On that year both Eids were after that date, and Founding day
        # holiday started at 2022; so what below works.
        self._observed_rule = (
            THU_TO_PREV_WED + FRI_TO_NEXT_SAT
            if self._year <= 2012
            else FRI_TO_PREV_THU + SAT_TO_NEXT_SUN
        )
        self.weekend = {THU, FRI} if self._year <= 2012 else {FRI, SAT}

        # Eid al-Fitr Holiday
        eid_al_fitr_name = tr("عطلة عيد الفطر")
        self._add_eid_al_fitr_day(eid_al_fitr_name)
        self._add_eid_al_fitr_day_two(eid_al_fitr_name)
        self._add_eid_al_fitr_day_three(eid_al_fitr_name)
        self._add_islamic_observed(self._add_eid_al_fitr_day_four(eid_al_fitr_name))

        # Arafat Day
        self._add_arafah_day(tr("يوم عرفة"))
        # Eid al-Adha Holiday
        name = tr("عطلة عيد الأضحى")
        self._add_eid_al_adha_day(name)
        self._add_eid_al_adha_day_two(name)
        self._add_islamic_observed(self._add_eid_al_adha_day_three(name))

        # If National Day happens within the Eid al-Fitr Holiday or
        # Eid al-Adha Holiday, there is no extra holidays given for it.
        if self._year >= 2005:
            dt = date(self._year, SEP, 23)
            if dt not in self:
                # National Day Holiday
                self._add_observed(self._add_holiday(tr("اليوم الوطني"), dt))

        # If Founding Day happens within the Eid al-Fitr Holiday or
        # Eid al-Adha Holiday, there is no extra holidays given for it.
        if self._year >= 2022:
            dt = date(self._year, FEB, 22)
            if dt not in self:
                # Founding Day
                self._add_observed(self._add_holiday(tr("يوم التأسيسي"), dt))


class SA(SaudiArabia):
    pass


class SAU(SaudiArabia):
    pass


class SaudiArabiaStaticHolidays:
    special_public_holidays = {
        # Celebrate the country's win against Argentina in the World Cup
        2022: (NOV, 23, tr("يوم وطني")),
    }

    special_public_holidays_observed = {
        # Eid al-Fitr Holiday
        2001: (JAN, 1, tr("عطلة عيد الفطر")),
    }



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see AUTHORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/python-holidays
#  License: MIT (see LICENSE file)

from typing import Tuple, Union

from holidays.calendars.gregorian import APR, MAY, JUN, JUL, SEP, DEC
from holidays.groups import ChristianHolidays, InternationalHolidays, StaticHolidays
from holidays.observed_holiday_base import (
    ObservedHolidayBase,
    MON_TO_NEXT_TUE,
    SAT_SUN_TO_NEXT_MON,
    SAT_SUN_TO_NEXT_MON_TUE,
)


class UnitedKingdom(ObservedHolidayBase, ChristianHolidays, InternationalHolidays, StaticHolidays):
    """
    References:
    - https://en.wikipedia.org/wiki/Public_holidays_in_the_United_Kingdom
    - https://archive.org/details/treatiseonbanki00walk/page/334/mode/2up
    - https://www.gov.uk/bank-holidays
    - https://www.timeanddate.com/holidays/uk/

    The Anniversary of the Battle of the Boyne bank holiday is proclaimed annually by the
    Secretary of State for Northern Ireland.

    In-Lieu observance was first provided in the Bank Holidays Extension Act 1875.
    """

    country = "GB"
    observed_label = "%s (observed)"
    subdivisions: Union[Tuple[()], Tuple[str, ...]] = (
        "ENG",  # England
        "NIR",  # Northern Ireland
        "SCT",  # Scotland
        "WLS",  # Wales
    )

    _deprecated_subdivisions: Tuple[str, ...] = (
        "England",
        "Northern Ireland",
        "Scotland",
        "UK",
        "Wales",
    )

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        StaticHolidays.__init__(self, UnitedKingdomStaticHolidays)
        kwargs.setdefault("observed_rule", SAT_SUN_TO_NEXT_MON)
        # Bank Holidays Extension Act 1875
        kwargs.setdefault("observed_since", 1875)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self) -> None:
        # Bank Holidays Act 1871
        if self._year <= 1871:
            return None

        # Good Friday
        self._add_good_friday("Good Friday")

        # May Day bank holiday (first Monday in May)
        if self._year >= 1978:
            name = "May Day"
            if self._year in {1995, 2020}:
                self._add_holiday_may_8(name)
            else:
                self._add_holiday_1st_mon_of_may(name)

        # Spring bank holiday (last Monday in May)
        if self._year >= 1971:
            spring_bank_dates = {
                2002: (JUN, 4),
                2012: (JUN, 4),
                2022: (JUN, 2),
            }
            name = "Spring Bank Holiday"
            if self._year in spring_bank_dates:
                self._add_holiday(name, spring_bank_dates[self._year])
            else:
                self._add_holiday_last_mon_of_may(name)

        if self.subdiv == "England":
            self._populate_subdiv_eng_public_holidays()
        elif self.subdiv == "Northern Ireland":
            self._populate_subdiv_nir_public_holidays()
        elif self.subdiv == "Scotland":
            self._populate_subdiv_sct_public_holidays()
        elif self.subdiv == "Wales":
            self._populate_subdiv_wls_public_holidays()

    def _populate_subdiv_holidays(self):
        # Bank Holidays Act 1871
        if self._year <= 1871:
            return None

        if self.subdiv not in {"SCT", "Scotland"}:
            # New Year's Day
            if self._year >= 1975:
                self._add_observed(self._add_new_years_day("New Year's Day"))

            # Christmas Day
            self._add_observed(
                self._add_christmas_day("Christmas Day"), rule=SAT_SUN_TO_NEXT_MON_TUE
            )

            # Boxing Day
            self._add_observed(
                self._add_christmas_day_two("Boxing Day"), rule=SAT_SUN_TO_NEXT_MON_TUE
            )

        super()._populate_subdiv_holidays()

    def _populate_subdiv_eng_public_holidays(self):
        # Easter Monday
        self._add_easter_monday("Easter Monday")

        # Whit Monday.
        if self._year <= 1970:
            self._add_whit_monday("Whit Monday")

        # Late Summer bank holiday (last Monday in August)
        if self._year >= 1971:
            self._add_holiday_last_mon_of_aug("Late Summer Bank Holiday")

    def _populate_subdiv_nir_public_holidays(self):
        if self._year >= 1903:
            # St. Patrick's Day
            self._add_observed(self._add_holiday_mar_17("St. Patrick's Day"))

        # Easter Monday
        self._add_easter_monday("Easter Monday")

        # Whit Monday.
        if self._year <= 1970:
            self._add_whit_monday("Whit Monday")

        # Battle of the Boyne
        self._add_observed(self._add_holiday_jul_12("Battle of the Boyne"))

        # Late Summer bank holiday (last Monday in August)
        if self._year >= 1971:
            self._add_holiday_last_mon_of_aug("Late Summer Bank Holiday")

    def _populate_subdiv_sct_public_holidays(self):
        # New Year's Day
        jan_1 = self._add_new_years_day("New Year's Day")

        # New Year Holiday
        self._add_observed(
            self._add_new_years_day_two("New Year Holiday"),
            rule=SAT_SUN_TO_NEXT_MON_TUE + MON_TO_NEXT_TUE,
        )
        self._add_observed(jan_1)

        # Summer bank holiday (first Monday in August)
        self._add_holiday_1st_mon_of_aug("Summer Bank Holiday")

        if self._year >= 2006:
            # St. Andrew's Day
            self._add_observed(self._add_holiday_nov_30("St. Andrew's Day"))

        # Christmas Day
        self._add_observed(
            self._add_christmas_day("Christmas Day"),
            rule=SAT_SUN_TO_NEXT_MON_TUE if self._year >= 1974 else SAT_SUN_TO_NEXT_MON,
        )

        if self._year >= 1974:
            # Boxing Day
            self._add_observed(
                self._add_christmas_day_two("Boxing Day"), rule=SAT_SUN_TO_NEXT_MON_TUE
            )

    def _populate_subdiv_wls_public_holidays(self):
        # Easter Monday
        self._add_easter_monday("Easter Monday")

        # Whit Monday.
        if self._year <= 1970:
            self._add_whit_monday("Whit Monday")

        # Late Summer bank holiday (last Monday in August)
        if self._year >= 1971:
            self._add_holiday_last_mon_of_aug("Late Summer Bank Holiday")


class UK(UnitedKingdom):
    pass


class GB(UnitedKingdom):
    pass


class GBR(UnitedKingdom):
    pass


class UnitedKingdomStaticHolidays:
    special_public_holidays = {
        1977: (JUN, 7, "Silver Jubilee of Elizabeth II"),
        1981: (JUL, 29, "Wedding of Charles and Diana"),
        1999: (DEC, 31, "Millennium Celebrations"),
        2002: (JUN, 3, "Golden Jubilee of Elizabeth II"),
        2011: (APR, 29, "Wedding of William and Catherine"),
        2012: (JUN, 5, "Diamond Jubilee of Elizabeth II"),
        2022: (
            (JUN, 3, "Platinum Jubilee of Elizabeth II"),
            (SEP, 19, "State Funeral of Queen Elizabeth II"),
        ),
        2023: (MAY, 8, "Coronation of Charles III"),
    }



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see AUTHORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.groups import ChristianHolidays, InternationalHolidays
from holidays.holiday_base import HolidayBase


class Armenia(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    Armenia holidays.

    References:
     - https://en.wikipedia.org/wiki/Public_holidays_in_Armenia
     - http://www.parliament.am/legislation.php?sel=show&ID=1274&lang=arm&enc=utf8
     - https://www.arlis.am/documentview.aspx?docid=259
    """

    country = "AM"
    default_language = "hy"
    supported_languages = ("en_US", "hy")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        if self._year <= 1990:
            return None

        # New Year's Day.
        name = tr("Նոր տարվա օր")
        self._add_new_years_day(name)
        self._add_new_years_day_two(name)

        # Christmas. Epiphany Day.
        self._add_holiday_jan_6(tr("Սուրբ Ծնունդ եւ Հայտնություն"))

        if 2010 <= self._year <= 2021:
            self._add_new_years_day_three(name)
            self._add_new_years_day_four(name)

            # Christmas Eve.
            self._add_holiday_jan_5(tr("Սուրբ Ծննդյան տոներ"))

            # The Day of Remembrance of the Dead.
            self._add_holiday_jan_7(tr("Մեռելոց հիշատակի օր"))

        if self._year >= 2003:
            # Army Day.
            self._add_holiday_jan_28(tr("Բանակի օր"))

        # Women's Day.
        self._add_womens_day(tr("Կանանց տոն"))

        if 1994 <= self._year <= 2001:
            # Motherhood and Beauty Day.
            self._add_holiday_apr_7(tr("Մայրության և գեղեցկության տոն"))

        # Armenian Genocide Remembrance Day.
        self._add_holiday_apr_24(tr("Եղեռնի զոհերի հիշատակի օր"))

        if self._year >= 2001:
            self._add_labor_day(
                # Labor Day.
                tr("Աշխատանքի օր")
                if self._year >= 2002
                # International Day of Workers' Solidarity.
                else tr("Աշխատավորների համերաշխության միջազգային օր")
            )

        if self._year >= 1995:
            # Victory and Peace Day.
            self._add_world_war_two_victory_day(tr("Հաղթանակի և Խաղաղության տոն"))

        # Republic Day.
        self._add_holiday_may_28(tr("Հանրապետության օր"))

        if self._year >= 1996:
            # Constitution Day.
            self._add_holiday_jul_5(tr("Սահմանադրության օր"))

        if self._year >= 1992:
            # Independence Day.
            self._add_holiday_sep_21(tr("Անկախության օր"))

        # New Year's Eve.
        self._add_new_years_eve(tr("Նոր տարվա գիշեր"))


class AM(Armenia):
    pass


class ARM(Armenia):
    pass



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see AUTHORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/python-holidays
#  License: MIT (see LICENSE file)

from holidays.groups import ChineseCalendarHolidays, InternationalHolidays
from holidays.observed_holiday_base import ObservedHolidayBase, SAT_SUN_TO_NEXT_WORKDAY


class Vietnam(ObservedHolidayBase, ChineseCalendarHolidays, InternationalHolidays):
    """
    https://publicholidays.vn/
    http://vbpl.vn/TW/Pages/vbpqen-toanvan.aspx?ItemID=11013 Article.115
    https://www.timeanddate.com/holidays/vietnam/
    """

    country = "VN"
    observed_label = "%s (observed)"

    def __init__(self, *args, **kwargs):
        ChineseCalendarHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        kwargs.setdefault("observed_rule", SAT_SUN_TO_NEXT_WORKDAY)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        dts_observed = set()

        # New Year's Day
        dts_observed.add(self._add_new_years_day("International New Year's Day"))

        # Lunar New Year
        self._add_chinese_new_years_eve("Vietnamese New Year's Eve")
        self._add_chinese_new_years_day("Vietnamese New Year")
        self._add_chinese_new_years_day_two("The second day of Tet Holiday")
        self._add_chinese_new_years_day_three("The third day of Tet Holiday")
        self._add_chinese_new_years_day_four("The forth day of Tet Holiday")
        self._add_chinese_new_years_day_five("The fifth day of Tet Holiday")

        # Vietnamese Kings' Commemoration Day
        # https://en.wikipedia.org/wiki/H%C3%B9ng_Kings%27_Festival
        if self._year >= 2007:
            dts_observed.add(self._add_hung_kings_day("Hung Kings Commemoration Day"))

        # Liberation Day/Reunification Day
        dts_observed.add(self._add_holiday_apr_30("Liberation Day/Reunification Day"))

        # International Labor Day
        dts_observed.add(self._add_labor_day("International Labor Day"))

        # Independence Day
        dts_observed.add(self._add_holiday_sep_2("Independence Day"))

        if self.observed:
            self._populate_observed(dts_observed)


class VN(Vietnam):
    pass


class VNM(Vietnam):
    pass



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see AUTHORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/python-holidays
#  License: MIT (see LICENSE file)

from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.groups import ChristianHolidays, InternationalHolidays, IslamicHolidays
from holidays.observed_holiday_base import ObservedHolidayBase, SAT_SUN_TO_NEXT_WORKDAY


class Kazakhstan(ObservedHolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays):
    """
    References:
        - https://www.officeholidays.com/countries/kazakhstan/2020
        - https://egov.kz/cms/en/articles/holidays-calend
        - https://en.wikipedia.org/wiki/Public_holidays_in_Kazakhstan
        - https://adilet.zan.kz/rus/docs/Z010000267%5F/history
    """

    country = "KZ"
    observed_label = "%s (observed)"

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self)
        kwargs.setdefault("observed_rule", SAT_SUN_TO_NEXT_WORKDAY)
        kwargs.setdefault("observed_since", 2002)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        # Kazakhstan declared its sovereignty on 25 October 1990
        if self._year <= 1990:
            return None

        dts_observed = set()

        # New Year's holiday (2 days)
        name = "New Year"
        dts_observed.add(self._add_new_years_day(name))
        dts_observed.add(self._add_new_years_day_two(name))

        # Orthodox Christmas (nonworking day, without extending)
        if self._year >= 2006:
            self._add_christmas_day("Orthodox Christmas")

        # International Women's Day
        dts_observed.add(self._add_womens_day("International Women's Day"))

        # Nauryz holiday
        if self._year >= 2002:
            name = "Nauryz holiday"
            dts_observed.add(self._add_holiday_mar_22(name))
            if self._year >= 2010:
                dts_observed.add(self._add_holiday_mar_21(name))
                dts_observed.add(self._add_holiday_mar_23(name))

        # Kazakhstan People Solidarity Holiday
        dts_observed.add(self._add_labor_day("Kazakhstan People Solidarity Holiday"))

        # Defender of the Fatherland Day
        if self._year >= 2013:
            dts_observed.add(self._add_holiday_may_7("Defender of the Fatherland Day"))

        # Victory Day
        dts_observed.add(self._add_world_war_two_victory_day("Victory Day"))

        # Capital Day
        if self._year >= 2009:
            dts_observed.add(self._add_holiday_jul_6("Capital Day"))

        # Constitution Day of the Republic of Kazakhstan
        if self._year >= 1996:
            dts_observed.add(
                self._add_holiday_aug_30("Constitution Day of the Republic of Kazakhstan")
            )

        # Republic Day
        if 1994 <= self._year <= 2008 or self._year >= 2022:
            dts_observed.add(self._add_holiday_oct_25("Republic Day"))

        # First President Day
        if 2012 <= self._year <= 2021:
            dts_observed.add(self._add_holiday_dec_1("First President Day"))

        # Kazakhstan Independence Day
        name = "Kazakhstan Independence Day"
        dts_observed.add(self._add_holiday_dec_16(name))
        if 2002 <= self._year <= 2021:
            dts_observed.add(self._add_holiday_dec_17(name))

        if self.observed:
            self._populate_observed(dts_observed)

        # Kurban Ait (nonworking day, without extending)
        if self._year >= 2006:
            self._add_eid_al_adha_day("Kurban Ait")


class KZ(Kazakhstan):
    pass


class KAZ(Kazakhstan):
    pass



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see AUTHORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars import _CustomIslamicHolidays
from holidays.calendars.gregorian import JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC
from holidays.constants import HALF_DAY, PUBLIC
from holidays.groups import InternationalHolidays, IslamicHolidays, StaticHolidays
from holidays.holiday_base import HolidayBase


class Turkey(HolidayBase, InternationalHolidays, IslamicHolidays, StaticHolidays):
    """
    References:

    - https://en.wikipedia.org/wiki/Public_holidays_in_Turkey
    - [Law 2739] https://www5.tbmm.gov.tr/tutanaklar/KANUNLAR_KARARLAR/kanuntbmmc015/kanuntbmmc015/kanuntbmmc01502739.pdf  # noqa: E501
    - [Law 2429] https://www.mevzuat.gov.tr/MevzuatMetin/1.5.2429.pdf
    - [Hijri calendar holidays] https://vakithesaplama.diyanet.gov.tr/hicriden_miladiye.php
    """

    country = "TR"
    default_language = "tr"
    # %s (estimated).
    estimated_label = tr("%s (tahmini)")
    supported_categories = (HALF_DAY, PUBLIC)
    supported_languages = ("en_US", "tr", "uk")

    def __init__(self, *args, **kwargs):
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self, TurkeyIslamicHolidays)
        StaticHolidays.__init__(self, TurkeyStaticHolidays)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        # Law 2739 of 27 May 1935.
        if self._year <= 1935:
            return None

        # New Year's Day.
        self._add_new_years_day(tr("Yılbaşı"))

        name = (
            # National Sovereignty and Children's Day.
            tr("Ulusal Egemenlik ve Çocuk Bayramı")
            if self._year >= 1981
            # National Sovereignty Day.
            else tr("Ulusal Egemenlik Bayramı")
        )
        self._add_holiday_apr_23(name)

        if self._year >= 2009:
            # Labour and Solidarity Day.
            self._add_labor_day(tr("Emek ve Dayanışma Günü"))

        name = (
            # Commemoration of Atatürk, Youth and Sports Day.
            tr("Atatürk'ü Anma, Gençlik ve Spor Bayramı")
            if self._year >= 1981
            # Youth and Sports Day.
            else tr("Gençlik ve Spor Bayramı")
        )
        self._add_holiday_may_19(name)

        if 1963 <= self._year <= 1980:
            # Freedom and Constitution Day.
            self._add_holiday_may_27(tr("Hürriyet ve Anayasa Bayramı"))

        if self._year >= 2017:
            # Democracy and National Unity Day.
            self._add_holiday_jul_15(tr("Demokrasi ve Millî Birlik Günü"))

        # Victory Day.
        self._add_holiday_aug_30(tr("Zafer Bayramı"))

        # Republic Day.
        name = tr("Cumhuriyet Bayramı")
        self._add_holiday_oct_29(name)
        if self._year <= 1980:
            self._add_holiday_oct_30(name)

        # Eid al-Fitr.
        name = tr("Ramazan Bayramı")
        self._add_eid_al_fitr_day(name)
        self._add_eid_al_fitr_day_two(name)
        self._add_eid_al_fitr_day_three(name)

        # Eid al-Adha.
        name = tr("Kurban Bayramı")
        self._add_eid_al_adha_day(name)
        self._add_eid_al_adha_day_two(name)
        self._add_eid_al_adha_day_three(name)
        self._add_eid_al_adha_day_four(name)

    def _populate_half_day_holidays(self):
        if self._year <= 1935:
            return None

        # %s (from 1pm).
        begin_time_label = self.tr("%s (saat 13.00'ten)")

        # Republic Day.
        self._add_holiday_oct_28(begin_time_label % self.tr("Cumhuriyet Bayramı"))

        # Eid al-Fitr.
        self._add_eid_al_fitr_eve(begin_time_label % self.tr("Ramazan Bayramı"))

        # Eid al-Adha.
        self._add_arafah_day(begin_time_label % self.tr("Kurban Bayramı"))


class TR(Turkey):
    pass


class TUR(Turkey):
    pass


class TurkeyIslamicHolidays(_CustomIslamicHolidays):
    EID_AL_ADHA_DATES = {
        1936: (MAR, 4),
        1937: (FEB, 22),
        1938: (FEB, 11),
        1939: (JAN, 31),
        1940: (JAN, 20),
        1941: ((JAN, 8), (DEC, 29)),
        1942: (DEC, 18),
        1943: (DEC, 8),
        1944: (NOV, 26),
        1945: (NOV, 15),
        1946: (NOV, 4),
        1947: (OCT, 25),
        1948: (OCT, 13),
        1949: (OCT, 3),
        1950: (SEP, 23),
        1951: (SEP, 12),
        1952: (AUG, 31),
        1953: (AUG, 20),
        1954: (AUG, 9),
        1955: (JUL, 30),
        1956: (JUL, 19),
        1957: (JUL, 8),
        1958: (JUN, 28),
        1959: (JUN, 17),
        1960: (JUN, 5),
        1961: (MAY, 25),
        1962: (MAY, 14),
        1963: (MAY, 4),
        1964: (APR, 23),
        1965: (APR, 12),
        1966: (APR, 1),
        1967: (MAR, 21),
        1968: (MAR, 10),
        1969: (FEB, 27),
        1970: (FEB, 17),
        1971: (FEB, 6),
        1972: (JAN, 27),
        1973: (JAN, 15),
        1974: ((JAN, 4), (DEC, 24)),
        1975: (DEC, 13),
        1976: (DEC, 2),
        1977: (NOV, 22),
        1978: (NOV, 11),
        1979: (OCT, 31),
        1980: (OCT, 19),
        1981: (OCT, 8),
        1982: (SEP, 27),
        1983: (SEP, 17),
        1984: (SEP, 6),
        1985: (AUG, 26),
        1986: (AUG, 16),
        1987: (AUG, 5),
        1988: (JUL, 24),
        1989: (JUL, 13),
        1990: (JUL, 3),
        1991: (JUN, 23),
        1992: (JUN, 11),
        1993: (JUN, 1),
        1994: (MAY, 21),
        1995: (MAY, 10),
        1996: (APR, 28),
        1997: (APR, 18),
        1998: (APR, 7),
        1999: (MAR, 28),
        2000: (MAR, 16),
        2001: (MAR, 5),
        2002: (FEB, 22),
        2003: (FEB, 11),
        2004: (FEB, 1),
        2005: (JAN, 20),
        2006: ((JAN, 10), (DEC, 31)),
        2007: (DEC, 20),
        2008: (DEC, 8),
        2009: (NOV, 27),
        2010: (NOV, 16),
        2011: (NOV, 6),
        2012: (OCT, 25),
        2013: (OCT, 15),
        2014: (OCT, 4),
        2015: (SEP, 24),
        2016: (SEP, 12),
        2017: (SEP, 1),
        2018: (AUG, 21),
        2019: (AUG, 11),
        2020: (JUL, 31),
        2021: (JUL, 20),
        2022: (JUL, 9),
        2023: (JUN, 28),
        2024: (JUN, 16),
        2025: (JUN, 6),
        2026: (MAY, 27),
        2027: (MAY, 16),
        2028: (MAY, 5),
        2029: (APR, 24),
        2030: (APR, 13),
        2031: (APR, 2),
        2032: (MAR, 22),
    }

    EID_AL_FITR_DATES = {
        1936: (DEC, 15),
        1937: (DEC, 4),
        1938: (NOV, 23),
        1939: (NOV, 13),
        1940: (NOV, 1),
        1941: (OCT, 22),
        1942: (OCT, 12),
        1943: (OCT, 1),
        1944: (SEP, 19),
        1945: (SEP, 8),
        1946: (AUG, 29),
        1947: (AUG, 18),
        1948: (AUG, 6),
        1949: (JUL, 27),
        1950: (JUL, 16),
        1951: (JUL, 6),
        1952: (JUN, 24),
        1953: (JUN, 13),
        1954: (JUN, 2),
        1955: (MAY, 23),
        1956: (MAY, 12),
        1957: (MAY, 1),
        1958: (APR, 20),
        1959: (APR, 9),
        1960: (MAR, 29),
        1961: (MAR, 18),
        1962: (MAR, 8),
        1963: (FEB, 25),
        1964: (FEB, 15),
        1965: (FEB, 3),
        1966: (JAN, 23),
        1967: (JAN, 12),
        1968: ((JAN, 1), (DEC, 21)),
        1969: (DEC, 11),
        1970: (DEC, 1),
        1971: (NOV, 20),
        1972: (NOV, 8),
        1973: (OCT, 28),
        1974: (OCT, 17),
        1975: (OCT, 6),
        1976: (SEP, 25),
        1977: (SEP, 15),
        1978: (SEP, 4),
        1979: (AUG, 24),
        1980: (AUG, 12),
        1981: (AUG, 1),
        1982: (JUL, 22),
        1983: (JUL, 12),
        1984: (JUN, 30),
        1985: (JUN, 20),
        1986: (JUN, 9),
        1987: (MAY, 29),
        1988: (MAY, 17),
        1989: (MAY, 6),
        1990: (APR, 26),
        1991: (APR, 16),
        1992: (APR, 4),
        1993: (MAR, 24),
        1994: (MAR, 13),
        1995: (MAR, 3),
        1996: (FEB, 20),
        1997: (FEB, 9),
        1998: (JAN, 29),
        1999: (JAN, 19),
        2000: ((JAN, 8), (DEC, 27)),
        2001: (DEC, 16),
        2002: (DEC, 5),
        2003: (NOV, 25),
        2004: (NOV, 14),
        2005: (NOV, 3),
        2006: (OCT, 23),
        2007: (OCT, 12),
        2008: (SEP, 30),
        2009: (SEP, 20),
        2010: (SEP, 9),
        2011: (AUG, 30),
        2012: (AUG, 19),
        2013: (AUG, 8),
        2014: (JUL, 28),
        2015: (JUL, 17),
        2016: (JUL, 5),
        2017: (JUN, 25),
        2018: (JUN, 15),
        2019: (JUN, 4),
        2020: (MAY, 24),
        2021: (MAY, 13),
        2022: (MAY, 2),
        2023: (APR, 21),
        2024: (APR, 10),
        2025: (MAR, 30),
        2026: (MAR, 20),
        2027: (MAR, 9),
        2028: (FEB, 26),
        2029: (FEB, 14),
        2030: (FEB, 4),
        2031: (JAN, 24),
        2032: (JAN, 14),
    }


class TurkeyStaticHolidays:
    special_public_holidays = {
        # Public holiday.
        1999: (DEC, 31, tr("Genel tati̇l"))
    }



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see AUTHORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import _get_all_sundays
from holidays.groups import ChristianHolidays, InternationalHolidays
from holidays.holiday_base import HolidayBase


class Sweden(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    Swedish holidays.
    Note that holidays falling on a sunday are "lost",
    it will not be moved to another day to make up for the collision.
    In Sweden, ALL sundays are considered a holiday
    (https://sv.wikipedia.org/wiki/Helgdagar_i_Sverige).
    Initialize this class with include_sundays=False
    to not include sundays as a holiday.

    Primary sources:
        - https://sv.wikipedia.org/wiki/Helgdagar_i_Sverige
        - http://www.riksdagen.se/sv/dokument-lagar/dokument/svensk-forfattningssamling/lag-1989253-om-allmanna-helgdagar_sfs-1989-253  # noqa: E501
    """

    country = "SE"
    default_language = "sv"
    supported_languages = ("en_US", "sv", "uk")

    def __init__(self, include_sundays=True, *args, **kwargs):
        """
        :param include_sundays:
            Whether to consider sundays as a holiday (which they are in Sweden)
        """
        self.include_sundays = include_sundays
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        # New Year's Day.
        self._add_new_years_day(tr("Nyårsdagen"))

        # Epiphany.
        self._add_epiphany_day(tr("Trettondedag jul"))

        if self._year <= 1953:
            # Feast of the Annunciation.
            self._add_holiday_mar_25(tr("Jungfru Marie bebådelsedag"))

        # Good Friday.
        self._add_good_friday(tr("Långfredagen"))

        # Easter Sunday.
        self._add_easter_sunday(tr("Påskdagen"))

        # Easter Monday.
        self._add_easter_monday(tr("Annandag påsk"))

        # Source: https://sv.wikipedia.org/wiki/F%C3%B6rsta_maj
        if self._year >= 1939:
            # May Day.
            self._add_labor_day(tr("Första maj"))

        # Ascension Day.
        self._add_ascension_thursday(tr("Kristi himmelsfärdsdag"))

        # Source: https://sv.wikipedia.org/wiki/Sveriges_nationaldag
        if self._year >= 2005:
            # National Day of Sweden.
            self._add_holiday_jun_6(tr("Sveriges nationaldag"))

        # Whit Sunday.
        self._add_whit_sunday(tr("Pingstdagen"))

        if self._year <= 2004:
            # Whit Monday.
            self._add_whit_monday(tr("Annandag pingst"))

        # Source:
        # https://sv.wikipedia.org/wiki/Midsommarafton
        # https://www.nordiskamuseet.se/aretsdagar/midsommarafton
        # Midsummer evening. Friday between June 19th and June 25th

        # Midsummer Eve.
        name = tr("Midsommarafton")
        dt = (
            self._add_holiday_1st_fri_from_jun_19(name)
            if self._year >= 1953
            else self._add_holiday_jun_23(name)
        )

        # Midsummer Day.
        self._add_holiday(tr("Midsommardagen"), dt + td(days=+1))

        # All Saints' Day.
        self._add_holiday_1st_sat_from_oct_31(tr("Alla helgons dag"))

        # Christmas Eve.
        self._add_christmas_eve(tr("Julafton"))

        # Christmas Day.
        self._add_christmas_day(tr("Juldagen"))

        # Second Day of Christmas.
        self._add_christmas_day_two(tr("Annandag jul"))

        # New Year's Eve.
        self._add_new_years_eve(tr("Nyårsafton"))

        # Optionally add all Sundays of the year.
        if self.include_sundays:
            for dt in _get_all_sundays(self._year):
                # Sunday.
                self._add_holiday(tr("Söndag"), dt)


class SE(Sweden):
    pass


class SWE(Sweden):
    pass



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see AUTHORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/python-holidays
#  License: MIT (see LICENSE file)

from holidays.calendars.gregorian import JAN, APR, JUN, AUG, SEP, OCT, DEC
from holidays.constants import BANK, HALF_DAY, PUBLIC
from holidays.groups import ChristianHolidays, InternationalHolidays, StaticHolidays
from holidays.observed_holiday_base import (
    ObservedHolidayBase,
    MON_TO_NEXT_TUE,
    SUN_TO_NEXT_MON,
    SUN_TO_NEXT_TUE,
    SAT_SUN_TO_NEXT_MON,
    SAT_SUN_TO_NEXT_MON_TUE,
)


class Australia(ObservedHolidayBase, ChristianHolidays, InternationalHolidays, StaticHolidays):
    """
    References:
      - https://www.qld.gov.au/recreation/travel/holidays
      - `ACT Holidays Act 1958 <https://www.legislation.act.gov.au/a/1958-19>`_
      - `ACT 2013-2023 <https://www.cmtedd.act.gov.au/archived-content/holidays/previous-years>`_
      - `NSW Banks and Bank Holidays Act 1912 <https://legislation.nsw.gov.au/view/html/repealed/current/act-1912-043>`_
      - `NSW Public Holidays Act 2010 <https://legislation.nsw.gov.au/view/html/inforce/current/act-2010-115>`_
      - `NT Public Holidays Act 1981 <https://legislation.nt.gov.au/api/sitecore/Act/PDF?id=12145>`_
      - `QLD Holidays Act 1983 <https://www.legislation.qld.gov.au/view/html/inforce/current/act-1983-018>`_
      - `QLD 2012-2024 <https://www.qld.gov.au/recreation/travel/holidays/public>`_
      - `SA Holidays Act 1910 <https://www.legislation.sa.gov.au/LZ/C/A/HOLIDAYS%20ACT%201910.aspx>`_
      - `SA Public Holidays Act 2023 <https://www.legislation.sa.gov.au/LZ/C/A/Public%20Holidays%20Act%202023.aspx>`_
      - `SA 2023-2024 <https://www.safework.sa.gov.au/resources/public-holidays>`_
      - `SA 2007-2021 <https://safework.sa.gov.au/__data/assets/pdf_file/0007/235474/Public-Holidays-since-2007.pdf>`_
      - `TAS Statutory Holidays Act 2000 <https://www.legislation.tas.gov.au/view/html/inforce/current/act-2000-096>`_
      - `VIC Public Holidays Act 1993 <https://www.legislation.vic.gov.au/in-force/acts/public-holidays-act-1993/027>`_
      - `VIC Minister appointment <https://www.gazette.vic.gov.au/gazette/Gazettes2015/GG2015S229.pdf>`_
      - `VIC 2018-2024 <https://business.vic.gov.au/business-information/public-holidays>`_
      - `WA Public and Bank Holidays Act 1972 <https://www.legislation.wa.gov.au/legislation/statutes.nsf/law_a639.html>`_
      - `WA 2019-2023 <https://www.commerce.wa.gov.au/labour-relations/previous-years-public-holiday-dates>`_
    """

    country = "AU"
    supported_categories = (BANK, HALF_DAY, PUBLIC)
    observed_label = "%s (observed)"
    subdivisions = ("ACT", "NSW", "NT", "QLD", "SA", "TAS", "VIC", "WA")
    subdivisions_aliases = {
        "Australian Capital Territory": "ACT",
        "New South Wales": "NSW",
        "Northern Territory": "NT",
        "Queensland": "QLD",
        "South Australia": "SA",
        "Tasmania": "TAS",
        "Victoria": "VIC",
        "Western Australia": "WA",
    }

    @property
    def sovereign_birthday(self) -> str:
        """Sovereign's birthday holiday name."""
        return (
            "King's Birthday"
            if 1902 <= self._year <= 1951 or self._year >= 2023
            else "Queen's Birthday"
        )

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        StaticHolidays.__init__(self, AustraliaStaticHolidays)
        kwargs.setdefault("observed_rule", SAT_SUN_TO_NEXT_MON)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        # Good Friday.
        self._add_good_friday("Good Friday")

        # Easter Monday.
        self._add_easter_monday("Easter Monday")

        # Sovereign's Birthday.
        if 1902 <= self._year <= 1935:
            if self._year >= 1912:
                self._add_holiday_jun_3(self.sovereign_birthday)  # George V
            else:
                self._add_holiday_nov_9(self.sovereign_birthday)  # Edward VII

        if self.subdiv:
            return None

        # Common holidays.

        # New Year's Day.
        self._add_new_years_day("New Year's Day")

        # Australia Day.
        if self._year >= 1935:
            self._add_holiday_jan_26("Australia Day")

        # ANZAC Day.
        if self._year >= 1921:
            self._add_anzac_day("ANZAC Day")

        # Christmas Day.
        self._add_christmas_day("Christmas Day")

        # Boxing Day.
        self._add_christmas_day_two("Boxing Day")

    def _populate_subdiv_act_public_holidays(self):
        # New Year's Day.
        # 1959-1992: SUN - move to MON.
        # 1993-2014: SAT, SUN - move to MON.
        # from 2015: SAT, SUN - add MON.

        # New Year's Day.
        name = "New Year's Day"
        if self._year >= 2015:
            self._add_observed(self._add_new_years_day(name))
        elif self._year >= 1959:
            jan_1 = (JAN, 1)
            if self._is_sunday(jan_1) or (self._year >= 1993 and self._is_saturday(jan_1)):
                self._add_holiday_1st_mon_from_jan_1(name)
            else:
                self._add_new_years_day(name)
        else:
            self._add_new_years_day(name)

        # Australia Day.
        # 1959-1989: not MON - move to MON.
        # from 1990: SAT, SUN - move to MON.

        # Australia Day.
        if self._year >= 1888:
            name = "Australia Day" if self._year >= 1935 else "Anniversary Day"
            if 1959 <= self._year <= 1989 or (self._year >= 1990 and self._is_weekend(JAN, 26)):
                self._add_holiday_1st_mon_from_jan_26(name)
            else:
                self._add_holiday_jan_26(name)

        # Canberra Day.
        if self._year >= 1913:
            name = "Canberra Day"
            if self._year <= 1958 or self._year == 2012:
                self._add_holiday_mar_12(name)
            elif self._year <= 2007:
                self._add_holiday_3rd_mon_of_mar(name)
            else:
                self._add_holiday_2nd_mon_of_mar(name)

        # Easter Saturday.
        self._add_holy_saturday("Easter Saturday")

        # Easter Sunday.
        if self._year >= 2016:
            self._add_easter_sunday("Easter Sunday")

        # ANZAC Day.
        # from 1959: SUN - move to MON.

        # ANZAC Day.
        if self._year >= 1921:
            name = "ANZAC Day"
            if self._year >= 1959 and self._is_sunday(APR, 25):
                self._add_holiday_1st_mon_from_apr_25(name)
            else:
                self._add_anzac_day(name)

        # Reconciliation Day.
        if self._year >= 2018:
            self._add_holiday_1st_mon_from_may_27("Reconciliation Day")

        # Sovereign's Birthday.
        if self._year >= 1936:
            self._add_holiday_2nd_mon_of_jun(self.sovereign_birthday)

        # Family & Community Day.
        # (First Monday of the September/October school holidays;
        # moved to the second Monday if this falls on Labour day).
        if 2010 <= self._year <= 2017:
            fc_dates = {
                2010: (SEP, 26),
                2011: (OCT, 10),
                2012: (OCT, 8),
                2013: (SEP, 30),
                2014: (SEP, 29),
                2015: (SEP, 28),
                2016: (SEP, 26),
                2017: (SEP, 25),
            }
            self._add_holiday("Family & Community Day", fc_dates[self._year])

        # Labour Day.
        self._add_holiday_1st_mon_of_oct("Labour Day")

        # Christmas Day.
        # 1958-1991: SUN - to MON.
        # 1992-2013: SAT, SUN - to MON.
        # from 2014: SAT - add MON, SUN - add TUE.

        # Christmas Day.
        name = "Christmas Day"
        if self._year >= 2014:
            self._add_observed(self._add_christmas_day(name), rule=SAT_SUN_TO_NEXT_MON_TUE)
        elif self._year >= 1958:
            dec_25 = (DEC, 25)
            if self._is_weekend(dec_25) or (self._year >= 1992 and self._is_saturday(dec_25)):
                self._add_holiday_1st_mon_from_dec_25(name)
            else:
                self._add_christmas_day(name)
        else:
            self._add_christmas_day(name)

        # Boxing Day.
        # 1958-1991: SUN - to MON, MON - to TUE.
        # 1992-2013: SAT - to MON, SUN - to TUE, MON - to TUE.
        # from 2014: SAT - add MON, SUN - add TUE.

        # Boxing Day.
        name = "Boxing Day"
        if self._year >= 2014:
            self._add_observed(self._add_christmas_day_two(name), rule=SAT_SUN_TO_NEXT_MON_TUE)
        elif self._year >= 1958:
            dec_26 = (DEC, 26)
            if self._year >= 1992:
                if self._is_saturday(dec_26):
                    self._add_holiday_1st_mon_from_dec_26(name)
                elif self._is_sunday(dec_26) or self._is_monday(dec_26):
                    self._add_holiday_1st_tue_from_dec_26(name)
                else:
                    self._add_christmas_day_two(name)
            else:
                if self._is_sunday(dec_26):
                    self._add_holiday_1st_mon_from_dec_26(name)
                elif self._is_monday(dec_26):
                    self._add_holiday_1st_tue_from_dec_26(name)
                else:
                    self._add_christmas_day_two(name)
        else:
            self._add_christmas_day_two(name)

    def _populate_subdiv_act_bank_holidays(self):
        # Bank Holiday.
        self._add_holiday_1st_mon_of_aug("Bank Holiday")

    def _populate_subdiv_nsw_public_holidays(self):
        # New Year's Day.
        # 1912-2010: SUN - add MON.
        # from 2011: SAT, SUN - add MON.

        # New Year's Day.
        self._add_observed(
            self._add_new_years_day("New Year's Day"),
            rule=SAT_SUN_TO_NEXT_MON if self._year >= 2011 else SUN_TO_NEXT_MON,
        )

        # Australia Day.
        # 1912-2010: SUN - add MON.
        # from 2011: SAT, SUN - to MON.

        # Australia Day.
        if self._year >= 1888:
            name = "Australia Day" if self._year >= 1946 else "Anniversary Day"
            if self._year >= 2011:
                if self._is_weekend(JAN, 26):
                    self._add_holiday_1st_mon_from_jan_26(name)
                else:
                    self._add_holiday_jan_26(name)
            else:
                self._add_observed(self._add_holiday_jan_26(name), rule=SUN_TO_NEXT_MON)

        # Easter Saturday.
        self._add_holy_saturday("Easter Saturday")

        # Easter Sunday.
        if self._year >= 2011:
            self._add_easter_sunday("Easter Sunday")

        # ANZAC Day.
        # 1912-2010: SUN - add MON.
        # from 2011: normal.

        # ANZAC Day.
        if self._year >= 1921:
            apr_25 = self._add_anzac_day("ANZAC Day")
            if self._year <= 2010:
                self._add_observed(apr_25)

        # Labour Day.
        self._add_holiday_1st_mon_of_oct("Labour Day")

        # Sovereign's Birthday.
        if self._year >= 1936:
            self._add_holiday_2nd_mon_of_jun(self.sovereign_birthday)

        # Bank Holiday.
        if 1912 <= self._year <= 2010:
            self._add_holiday_1st_mon_of_aug("Bank Holiday")

        # Christmas Day.
        # 1912-2010: SUN - add TUE.
        # from 2011: SAT - add MON, SUN - add TUE.

        # Christmas Day.
        self._add_observed(
            self._add_christmas_day("Christmas Day"),
            rule=SAT_SUN_TO_NEXT_MON_TUE if self._year >= 2011 else SUN_TO_NEXT_TUE,
        )

        # Boxing Day.
        # 1912-2010: SUN - add MON.
        # from 2011: SAT - add MON, SUN - add TUE.

        # Boxing Day.
        self._add_observed(
            self._add_christmas_day_two("Boxing Day"),
            rule=SAT_SUN_TO_NEXT_MON_TUE if self._year >= 2011 else SUN_TO_NEXT_MON,
        )

    def _populate_subdiv_nsw_bank_holidays(self):
        # Bank Holiday.
        if self._year >= 2011:
            self._add_holiday_1st_mon_of_aug("Bank Holiday")

    def _populate_subdiv_nt_public_holidays(self):
        # New Year's Day.
        # 1982-2016: SAT, SUN - move to MON.
        # from 2017: SAT, SUN - add MON.

        # New Year's Day.
        name = "New Year's Day"
        if self._year >= 2017:
            self._add_observed(self._add_new_years_day(name))
        elif self._year >= 1982 and self._is_weekend(JAN, 1):
            self._add_holiday_1st_mon_from_jan_1(name)
        else:
            self._add_new_years_day(name)

        # Australia Day.
        # from 1982: SAT, SUN - move to MON.

        # Australia Day.
        if self._year >= 1888:
            name = "Australia Day" if self._year >= 1935 else "Anniversary Day"
            if self._year >= 1982 and self._is_weekend(JAN, 26):
                self._add_holiday_1st_mon_from_jan_26(name)
            else:
                self._add_holiday_jan_26(name)

        # Easter Saturday.
        self._add_holy_saturday("Easter Saturday")

        # Easter Sunday.
        if self._year >= 2024:
            self._add_easter_sunday("Easter Sunday")

        # ANZAC Day.
        # from 1982: SUN - to MON.

        # ANZAC Day.
        if self._year >= 1921:
            name = "ANZAC Day"
            if self._year >= 1982 and self._is_sunday(APR, 25):
                self._add_holiday_1st_mon_from_apr_25(name)
            else:
                self._add_anzac_day(name)

        # Labour Day.
        self._add_holiday_1st_mon_of_may("May Day")

        # Sovereign's Birthday.
        if self._year >= 1936:
            self._add_holiday_2nd_mon_of_jun(self.sovereign_birthday)

        # Picnic Day.
        self._add_holiday_1st_mon_of_aug("Picnic Day")

        # Christmas Day.
        # 1981-2015: SAT, SUN - to MON.
        # from 2016: SAT, SUN - add MON.

        # Christmas Day.
        name = "Christmas Day"
        if self._year >= 2016:
            self._add_observed(self._add_christmas_day(name))
        elif self._year >= 1981 and self._is_weekend(DEC, 25):
            self._add_holiday_1st_mon_from_dec_25(name)
        else:
            self._add_christmas_day(name)

        # Boxing Day.
        # 1981-2022: SAT - to MON, SUN - to TUE, MON - to TUE.
        # from 2023: SAT - add MON, SUN - add TUE, MON - add TUE.
        name = "Boxing Day"
        if self._year >= 2023:
            self._add_observed(
                self._add_christmas_day_two(name), rule=SAT_SUN_TO_NEXT_MON_TUE + MON_TO_NEXT_TUE
            )
        elif self._year >= 1981:
            dec_26 = (DEC, 26)
            if self._is_saturday(dec_26):
                self._add_holiday_1st_mon_from_dec_26(name)
            elif self._is_sunday(dec_26) or self._is_monday(dec_26):
                self._add_holiday_1st_tue_from_dec_26(name)
            else:
                self._add_christmas_day_two(name)
        else:
            self._add_christmas_day_two(name)

    def _populate_subdiv_nt_half_day_holidays(self):
        if self._year >= 2016:
            begin_time_label = "%s (from 7pm)"

            # Christmas Eve.
            self._add_christmas_eve(begin_time_label % "Christmas Eve")

            # New Year's Eve.
            self._add_new_years_eve(begin_time_label % "New Year's Eve")

    def _populate_subdiv_qld_public_holidays(self):
        # New Year's Day.
        # 1984-2011: SUN - move to MON.
        # from 2012: SAT, SUN - add MON.

        # New Year's Day.
        name = "New Year's Day"
        if self._year >= 2012:
            self._add_observed(self._add_new_years_day(name))
        elif self._year >= 1984 and self._is_sunday(JAN, 1):
            self._add_holiday_1st_mon_from_jan_1(name)
        else:
            self._add_new_years_day(name)

        # Australia Day.
        # 1984-1995: not MON - move to MON.
        # from 1996: SAT, SUN - move to MON.

        # Australia Day.
        if self._year >= 1888:
            name = "Australia Day" if self._year >= 1935 else "Anniversary Day"
            if 1984 <= self._year <= 1995 or (self._year >= 1996 and self._is_weekend(JAN, 26)):
                self._add_holiday_1st_mon_from_jan_26(name)
            else:
                self._add_holiday_jan_26(name)

        # Easter Saturday.
        self._add_holy_saturday("Easter Saturday")

        # Easter Sunday.
        if self._year >= 2017:
            self._add_easter_sunday("Easter Sunday")

        # ANZAC Day.
        # from 1984: SUN - move to MON.

        # ANZAC Day.
        if self._year >= 1921:
            name = "ANZAC Day"
            if self._year >= 1984 and self._is_sunday(APR, 25):
                self._add_holiday_1st_mon_from_apr_25(name)
            else:
                self._add_anzac_day(name)

        # Labour Day.
        name = "Labour Day"
        if 2013 <= self._year <= 2015:
            self._add_holiday_1st_mon_of_oct(name)
        else:
            self._add_holiday_1st_mon_of_may(name)

        # Sovereign's Birthday.
        if self._year >= 1936:
            if self._year <= 2015 and self._year != 2012:
                self._add_holiday_2nd_mon_of_jun(self.sovereign_birthday)
            else:
                self._add_holiday_1st_mon_of_oct(self.sovereign_birthday)

        # The Royal Queensland Show (Ekka).
        # The Show starts on the first Friday of August - providing this is
        # not prior to the 5th - in which case it will begin on the second
        # Friday. The Wednesday during the show is a public holiday.
        ekka_dates = {
            2020: (AUG, 14),
            2021: (OCT, 29),
        }
        name = "The Royal Queensland Show"
        if self._year in ekka_dates:
            self._add_holiday(name, ekka_dates[self._year])
        else:
            # [1st FRI after Aug 5] + 5 days = [1st WED after Aug 10]
            self._add_holiday_1st_wed_from_aug_10(name)

        # Christmas Day.
        # 1984-2010: SUN - to MON.
        # from 2011: SAT - add MON, SUN - add TUE.

        # Christmas Day.
        name = "Christmas Day"
        if self._year >= 2011:
            self._add_observed(self._add_christmas_day(name), rule=SAT_SUN_TO_NEXT_MON_TUE)
        elif self._year >= 1984 and self._is_sunday(DEC, 25):
            self._add_holiday_1st_mon_from_dec_25(name)
        else:
            self._add_christmas_day(name)

        # Boxing Day.
        # 1984-1910: SUN - to MON, MON - to TUE.
        # from 2011: SAT - add MON, SUN - add TUE.

        # Boxing Day.
        name = "Boxing Day"
        if self._year >= 2011:
            self._add_observed(self._add_christmas_day_two(name), rule=SAT_SUN_TO_NEXT_MON_TUE)
        elif self._year >= 1984:
            dec_26 = (DEC, 26)
            if self._is_sunday(dec_26):
                self._add_holiday_1st_mon_from_dec_26(name)
            elif self._is_monday(dec_26):
                self._add_holiday_1st_tue_from_dec_26(name)
            else:
                self._add_christmas_day_two(name)
        else:
            self._add_christmas_day_two(name)

    def _populate_subdiv_sa_public_holidays(self):
        # New Year's Day.
        # 1984-2003: SAT, SUN - move to MON.
        # 2004-2023: SAT - move to MON, SUN - add MON.
        # from 2024: SAT, SUN - add MON.

        # New Year's Day.
        name = "New Year's Day"
        jan_1 = (JAN, 1)
        if self._year >= 2024:
            self._add_observed(self._add_new_years_day(name))
        elif self._year >= 2004:
            if self._is_saturday(jan_1):
                self._add_holiday_1st_mon_from_jan_1(name)
            else:
                self._add_observed(self._add_new_years_day(name), rule=SUN_TO_NEXT_MON)
        elif self._year >= 1984 and self._is_weekend(jan_1):
            self._add_holiday_1st_mon_from_jan_1(name)
        else:
            self._add_new_years_day(name)

        # Australia Day.
        # 1984-1993: not MON - move to MON.
        # 1994-2003: SAT, SUN - move to MON.
        # 2004-2023: SAT - move to MON, SUN - add MON.
        # from 2024: SAT, SUN - move to MON.

        # Australia Day.
        if self._year >= 1935:
            name = "Australia Day"
            jan_26 = (JAN, 26)
            if 1994 <= self._year <= 2003 or self._year >= 2024:
                if self._is_weekend(jan_26):
                    self._add_holiday_1st_mon_from_jan_26(name)
                else:
                    self._add_holiday_jan_26(name)
            elif self._year >= 2004:
                if self._is_saturday(jan_26):
                    self._add_holiday_1st_mon_from_jan_26(name)
                else:
                    self._add_observed(self._add_holiday_jan_26(name), rule=SUN_TO_NEXT_MON)
            elif self._year >= 1984:
                self._add_holiday_1st_mon_from_jan_26(name)
            else:
                self._add_holiday_jan_26(name)

        # Adelaide Cup Day.
        # 2006-2023: changed each year by SA Government Proclamation from the 3rd Monday in May
        # to the 2nd Monday in March.
        # from 2024: changed to the 2nd Monday in March officially.

        # Adelaide Cup Day.
        name = "Adelaide Cup Day"
        if self._year >= 2006:
            self._add_holiday_2nd_mon_of_mar(name)
        else:
            self._add_holiday_3rd_mon_of_may(name)

        # Easter Saturday.
        self._add_holy_saturday("Easter Saturday")

        # Easter Sunday.
        if self._year >= 2024:
            self._add_easter_sunday("Easter Sunday")

        # ANZAC Day.
        if self._year >= 1921:
            apr_25 = self._add_anzac_day("ANZAC Day")
            if self._year <= 2023:
                self._add_observed(apr_25, rule=SUN_TO_NEXT_MON)

        # Sovereign's Birthday.
        if self._year >= 1936:
            self._add_holiday_2nd_mon_of_jun(self.sovereign_birthday)

        # Labour Day.
        self._add_holiday_1st_mon_of_oct("Labour Day")

        # Proclamation Day.
        # 1984-1992: SAT, SUN - move to MON.
        # 1993-2002: SAT - to MON, SUN - to TUE, MON - to TUE.
        # 2003-2023: SAT - to MON, SUN - add TUE, MON - add TUE.
        # from 2024: SAT - add MON, SUN - add TUE, MON - add TUE.
        # (Placed before Christmas Day for proper observed calculation).

        # Proclamation Day.
        name = "Proclamation Day"
        dec_26 = (DEC, 26)
        if self._year >= 2024:
            self._add_observed(
                self._add_christmas_day_two(name), rule=SAT_SUN_TO_NEXT_MON_TUE + MON_TO_NEXT_TUE
            )
        elif self._year >= 2003:
            if self._is_saturday(dec_26):
                self._add_holiday_1st_mon_from_dec_26(name)
            else:
                self._add_observed(
                    self._add_christmas_day_two(name), rule=SUN_TO_NEXT_TUE + MON_TO_NEXT_TUE
                )
        elif self._year >= 1993:
            if self._is_saturday(dec_26):
                self._add_holiday_1st_mon_from_dec_26(name)
            elif self._is_sunday(dec_26) or self._is_monday(dec_26):
                self._add_holiday_1st_tue_from_dec_26(name)
            else:
                self._add_christmas_day_two(name)
        elif self._year >= 1984 and self._is_weekend(dec_26):
            self._add_holiday_1st_mon_from_dec_26(name)
        else:
            self._add_christmas_day_two(name)

        # Christmas Day.
        # 1984-2002: SAT, SUN - move to MON.
        # 2003-2023: SAT - move to MON, SUN - add MON.
        # from 2024: SAT, SUN - add MON.

        # Christmas Day.
        name = "Christmas Day"
        dec_25 = (DEC, 25)
        if self._year >= 2024:
            self._add_observed(self._add_christmas_day(name))
        elif self._year >= 2003:
            if self._is_saturday(dec_25):
                self._add_holiday_1st_mon_from_dec_25(name)
            else:
                self._add_observed(self._add_christmas_day(name), rule=SUN_TO_NEXT_MON)
        elif self._year >= 1984 and self._is_weekend(dec_25):
            self._add_holiday_1st_mon_from_dec_25(name)
        else:
            self._add_christmas_day(name)

    def _populate_subdiv_sa_half_day_holidays(self):
        if self._year >= 2012:
            begin_time_label = "%s (from 7pm)"

            # Christmas Eve.
            self._add_christmas_eve(begin_time_label % "Christmas Eve")

            # New Year's Eve.
            self._add_new_years_eve(begin_time_label % "New Year's Eve")

    def _populate_subdiv_tas_public_holidays(self):
        # New Year's Day.
        # from 2001: SAT, SUN - move to MON.

        # New Year's Day.
        name = "New Year's Day"
        if self._year >= 2001 and self._is_weekend(JAN, 1):
            self._add_holiday_1st_mon_from_jan_1(name)
        else:
            self._add_new_years_day(name)

        # Australia Day.
        # from 2001: SAT, SUN - move to MON.

        # Australia Day.
        if self._year >= 1888:
            name = "Australia Day" if self._year >= 1935 else "Anniversary Day"
            if self._year >= 2001 and self._is_weekend(JAN, 26):
                self._add_holiday_1st_mon_from_jan_26(name)
            else:
                self._add_holiday_jan_26(name)

        # Labour Day.
        self._add_holiday_2nd_mon_of_mar("Eight Hours Day")

        # Easter Tuesday.
        if self._year <= 2010:
            self._add_holiday_2_days_past_easter("Easter Tuesday")

        # ANZAC Day.
        if self._year >= 1921:
            self._add_anzac_day("ANZAC Day")

        # Sovereign's Birthday.
        if self._year >= 1936:
            self._add_holiday_2nd_mon_of_jun(self.sovereign_birthday)

        # Christmas Day.
        # 2000-2009: SAT - to MON, SUN - to TUE.
        # from 2010: SAT - add MON, SUN - add TUE.

        # Christmas Day.
        name = "Christmas Day"
        if self._year >= 2010:
            self._add_observed(self._add_christmas_day(name), rule=SAT_SUN_TO_NEXT_MON_TUE)
        elif self._year >= 2000:
            dec_25 = (DEC, 25)
            if self._is_saturday(dec_25):
                self._add_holiday_1st_mon_from_dec_25(name)
            elif self._is_sunday(dec_25):
                self._add_holiday_1st_tue_from_dec_25(name)
            else:
                self._add_christmas_day(name)
        else:
            self._add_christmas_day(name)

        # Boxing Day.
        # from 2000: SAT - to MON, SUN - to TUE.

        # Boxing Day.
        name = "Boxing Day"
        if self._year >= 2000:
            dec_26 = (DEC, 26)
            if self._is_saturday(dec_26):
                self._add_holiday_1st_mon_from_dec_26(name)
            elif self._is_sunday(dec_26):
                self._add_holiday_1st_tue_from_dec_26(name)
            else:
                self._add_christmas_day_two(name)
        else:
            self._add_christmas_day_two(name)

    def _populate_subdiv_vic_public_holidays(self):
        # New Year's Day.
        # 1994-1997: SUN - add MON.
        # 1998-2008: SUN - move to MON.
        # from 2009: SAT, SUN - add MON.

        # New Year's Day.
        name = "New Year's Day"
        if self._year >= 2009:
            self._add_observed(self._add_new_years_day(name))
        elif self._year >= 1998:
            if self._is_sunday(JAN, 1):
                self._add_holiday_1st_mon_from_jan_1(name)
            else:
                self._add_new_years_day(name)
        elif self._year >= 1994:
            self._add_observed(self._add_new_years_day(name), rule=SUN_TO_NEXT_MON)
        else:
            self._add_new_years_day(name)

        # Australia Day.
        # from 2009: SAT, SUN - move to MON.

        # Australia Day.
        if self._year >= 1888:
            name = "Australia Day" if self._year >= 1935 else "Anniversary Day"
            if self._year >= 2009 and self._is_weekend(JAN, 26):
                self._add_holiday_1st_mon_from_jan_26(name)
            else:
                self._add_holiday_jan_26(name)

        # Labour Day.
        self._add_holiday_2nd_mon_of_mar("Labour Day")

        # Easter Saturday.
        if self._year >= 2003:
            self._add_holy_saturday("Easter Saturday")

        # Easter Sunday.
        if self._year >= 2016:
            self._add_easter_sunday("Easter Sunday")

        # ANZAC Day.
        if self._year >= 1921:
            self._add_anzac_day("ANZAC Day")

        # Sovereign's Birthday.
        if self._year >= 1936:
            self._add_holiday_2nd_mon_of_jun(self.sovereign_birthday)

        # Grand Final Day.
        if self._year >= 2015:
            grand_final_dates = {
                2015: (OCT, 2),
                2016: (SEP, 30),
                # Rescheduled due to COVID-19.
                2020: (OCT, 23),
            }
            name = "Grand Final Day"
            if self._year in grand_final_dates:
                self._add_holiday(name, grand_final_dates[self._year])
            else:
                self._add_holiday_1_day_prior_last_sat_of_sep(name)

        # Melbourne Cup Day.
        if self._year >= 2009:
            self._add_holiday_1st_tue_of_nov("Melbourne Cup Day")

        # Christmas Day.
        # 2008-2018: SAT - to MON, SUN - to TUE.
        # from 2019: SAT - add MON, SUN - add TUE.

        # Christmas Day.
        name = "Christmas Day"
        if self._year >= 2019:
            self._add_observed(self._add_christmas_day(name), rule=SAT_SUN_TO_NEXT_MON_TUE)
        elif self._year >= 2000:
            dec_25 = (DEC, 25)
            if self._is_saturday(dec_25):
                self._add_holiday_1st_mon_from_dec_25(name)
            elif self._is_sunday(dec_25):
                self._add_holiday_1st_tue_from_dec_25(name)
            else:
                self._add_christmas_day(name)
        else:
            self._add_christmas_day(name)

        # Boxing Day.
        # 1994-2008: SUN - to MON.
        # from 2009: SAT - add MON, SUN - add TUE.

        # Boxing Day.
        name = "Boxing Day"
        if self._year >= 2009:
            self._add_observed(self._add_christmas_day_two(name), rule=SAT_SUN_TO_NEXT_MON_TUE)
        elif self._year >= 1994 and self._is_sunday(DEC, 26):
            self._add_holiday_1st_tue_from_dec_26(name)
        else:
            self._add_christmas_day_two(name)

    def _populate_subdiv_wa_public_holidays(self):
        # New Year's Day.
        # from 1973: SAT, SUN - add MON.

        # New Year's Day.
        jan_1 = self._add_new_years_day("New Year's Day")
        if self._year >= 1973:
            self._add_observed(jan_1)

        # Australia Day.
        # 1973-1993: not MON - move to MON.
        # from 1994: SAT, SUN - move to MON.

        # Australia Day.
        if self._year >= 1888:
            name = "Australia Day" if self._year >= 1935 else "Anniversary Day"
            if self._year >= 1994:
                if self._is_weekend(JAN, 26):
                    self._add_holiday_1st_mon_from_jan_26(name)
                else:
                    self._add_holiday_jan_26(name)
            elif self._year >= 1973:
                self._add_holiday_1st_mon_from_jan_26(name)
            else:
                self._add_holiday_jan_26(name)

        # Labour Day.
        self._add_holiday_1st_mon_of_mar("Labour Day")

        # Easter Sunday.
        if self._year >= 2022:
            self._add_easter_sunday("Easter Sunday")

        # ANZAC Day.
        if self._year >= 1921:
            apr_25 = self._add_anzac_day("ANZAC Day")
            if self._year >= 1973:
                self._add_observed(apr_25)

        # Western Australia Day.
        if self._year >= 1833:
            self._add_holiday_1st_mon_of_jun(
                "Western Australia Day" if self._year >= 2012 else "Foundation Day"
            )

        # Sovereign's Birthday.
        if self._year >= 1936:
            if self._year >= 1984:
                # Celebration Day for the Anniversary of the Birthday of the Reigning Sovereign
                # to be appointed for each year by proclamation published in the Government Gazette
                sovereign_birthday_dates = {
                    2011: (OCT, 28),
                    2012: (OCT, 1),
                    2024: (SEP, 23),
                }
                if self._year in sovereign_birthday_dates:
                    self._add_holiday(
                        self.sovereign_birthday, sovereign_birthday_dates[self._year]
                    )
                else:
                    self._add_holiday_last_mon_of_sep(self.sovereign_birthday)
            else:
                self._add_holiday_2nd_mon_of_oct(self.sovereign_birthday)

        # Boxing Day.
        # 1972-1975: SAT - add MON, SUN - add TUE.
        # from 1976: SAT - add MON, SUN - add TUE, MON - add TUE.
        # (Placed before Christmas Day for proper observed calculation).

        # Boxing Day.
        dec_26 = self._add_christmas_day_two("Boxing Day")
        if self._year >= 1972:
            self._add_observed(
                dec_26,
                rule=SAT_SUN_TO_NEXT_MON_TUE + MON_TO_NEXT_TUE
                if self._year >= 1976
                else SAT_SUN_TO_NEXT_MON_TUE,
            )

        # Christmas Day.
        # from 1972: SAT, SUN - add MON.

        # Christmas Day.
        dec_25 = self._add_christmas_day("Christmas Day")
        if self._year >= 1972:
            self._add_observed(dec_25)


class AU(Australia):
    pass


class AUS(Australia):
    pass


class AustraliaStaticHolidays:
    special_public_holidays = {
        2022: (SEP, 22, "National Day of Mourning for Queen Elizabeth II"),
    }

    special_act_public_holidays = {
        2020: (APR, 20, "Declared public holiday"),
        2021: (APR, 25, "Additional public holiday"),
    }

    special_qld_public_holidays = {
        2012: (JUN, 11, "Queen's Diamond Jubilee"),
    }

    special_qld_public_holidays_observed = {
        2010: (DEC, 28, "Christmas Day"),
        2011: (JAN, 3, "New Year's Day"),
    }

    special_wa_public_holidays = {
        # In 2011 both ANZAC Day and Easter Monday fell on Monday 25 April.
        2011: (APR, 26, "Special public holiday"),
    }



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see AUTHORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.groups import IslamicHolidays, PersianCalendarHolidays
from holidays.holiday_base import HolidayBase


class Iran(HolidayBase, IslamicHolidays, PersianCalendarHolidays):
    """
    References:
    - https://en.wikipedia.org/wiki/Public_holidays_in_Iran
    - https://fa.wikipedia.org/wiki/تعطیلات_عمومی_در_ایران
    """

    country = "IR"
    default_language = "fa"
    # %s (estimated).
    estimated_label = tr("(تخمین زده) %s")
    supported_languages = ("en_US", "fa")

    def __init__(self, *args, **kwargs):
        IslamicHolidays.__init__(self)
        PersianCalendarHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        if self._year <= 1979:
            return None

        # Persian New Year.
        name = tr("نوروز")
        self._add_nowruz_day(name)
        self._add_nowruz_day_two(name)
        self._add_nowruz_day_three(name)
        self._add_nowruz_day_four(name)

        # Islamic Republic Day.
        self._add_islamic_republic_day(tr("روز جمهوری اسلامی"))

        # Nature's Day.
        self._add_natures_day(tr("روز طبیعت"))

        # Death of Khomeini.
        self._add_death_of_khomeini_day(tr("درگذشت سید روح‌الله خمینی"))

        # Khordad National Uprising.
        self._add_khordad_uprising_day(tr("تظاهرات ۱۵ خرداد"))

        # Islamic Revolution Day.
        self._add_islamic_revolution_day(tr("پیروزی انقلاب ۵۷"))

        # Iranian Oil Industry Nationalization Day.
        self._add_oil_nationalization_day(tr("ملی‌شدن صنعت نفت"))

        # Tasua.
        self._add_tasua_day(tr("تاسوعا"))

        # Ashura.
        self._add_ashura_day(tr("کشته‌شدن حسین بن علی، عاشورا"))

        # Arbaeen.
        self._add_arbaeen_day(tr("چهلم حسین بن علی اربعین"))

        # Demise of Prophet Muhammad and Hasan ibn Ali.
        self._add_prophet_death_day(tr("کشته‌شدن حسن مجتبی و درگذشت محمد"))

        # Martyrdom of Ali al-Rida.
        self._add_ali_al_rida_death_day(tr("کشته‌شدن علی بن موسی الرضا"))

        # Martyrdom of Hasan al-Askari.
        self._add_hasan_al_askari_death_day(tr("کشته‌شدن حسن عسکری"))

        # Birthday of Muhammad and Ja'far al-Sadiq.
        self._add_sadiq_birthday_day(tr("زادروز محمد و جعفر صادق"))

        # Martyrdom of Fatima.
        self._add_fatima_death_day(tr("کشته‌شدن فاطمه زهرا"))

        # Birthday of Ali.
        self._add_ali_birthday_day(tr("زادروز علی بن ابی‌طالب"))

        # Ascension of Muhammad.
        self._add_isra_and_miraj_day(tr("مبعث"))

        # Birthday of Mahdi.
        self._add_imam_mahdi_birthday_day(tr("زادروز حجت بن الحسن"))

        # Martyrdom of Ali.
        self._add_ali_death_day(tr("کشته‌شدن علی بن ابی‌طالب"))

        # Eid al-Fitr.
        name = tr("عید فطر")
        self._add_eid_al_fitr_day(name)
        self._add_eid_al_fitr_day_two(name)

        # Martyrdom of Ja'far al-Sadiq.
        self._add_sadiq_death_day(tr("کشته‌شدن جعفر صادق"))

        # Eid al-Adha.
        self._add_eid_al_adha_day(tr("عید قربان"))

        # Eid al-Ghadeer.
        self._add_eid_al_ghadir_day(tr("عید غدیر"))


class IR(Iran):
    pass


class IRN(Iran):
    pass



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see AUTHORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/python-holidays
#  License: MIT (see LICENSE file)

from holidays.calendars.gregorian import SEP, NOV
from holidays.constants import ARMED_FORCES, HALF_DAY, PUBLIC
from holidays.groups import ChristianHolidays, InternationalHolidays, StaticHolidays
from holidays.observed_holiday_base import ObservedHolidayBase, SAT_TO_PREV_FRI, SUN_TO_NEXT_MON


class Palau(ObservedHolidayBase, ChristianHolidays, InternationalHolidays):
    """
    References:
    - http://www.paclii.org/pw/legis/consol_act/gpt1262/  # Chapter 7, Holidays.
    - https://www.palaugov.pw/wp-content/uploads/2017/11/RPPL-No.-10-15-re.-Family-Day-Holiday.pdf
    - https://www.facebook.com/PalauPresident/posts/195883107230463 # EO336 Memorial Day repealed
    - https://www.taiwanembassy.org/pal_en/post/792.html # Earliest source for President's Day

    If any of the holidays enumerated in section 701 of this chapter falls on Sunday, the
    following Monday shall be observed as a holiday. If any of the holidays enumerated in
    section 701 of this chapter falls on Saturday, the preceding Friday shall be observed
    as a holiday.

    As there's no record of President's Day (Jun 1) and Independence Day (Oct 1) being
    legal holiday before 2017, as seen in RPRL 10-15, they shall be assumed to start in 2018
    for our current implementation.
    """

    country = "PW"
    supported_categories = (ARMED_FORCES, HALF_DAY, PUBLIC)
    observed_label = "%s (observed)"

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        StaticHolidays.__init__(self, PalauStaticHolidays)
        kwargs.setdefault("observed_rule", SAT_TO_PREV_FRI + SUN_TO_NEXT_MON)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        # Republic of Palau Public Law No. 2-15.
        # The legislation was first adopted by the 2nd Olbiil Era Kelulau (1984-1988),
        # but since we cannot find any info on its actual adoption date, we may as
        # well use the formation date of the country as the placeholder cut-off date.
        if self._year <= 1980:
            return None

        # Fixed Date Public Holidays.

        # New Year's Day.
        name = "New Year's Day"
        self._add_observed(self._add_new_years_day(name))
        self._add_observed(self._next_year_new_years_day, name=name, rule=SAT_TO_PREV_FRI)

        # Youth Day.
        self._add_observed(self._add_holiday_mar_15("Youth Day"))

        # Senior Citizens Day.
        self._add_observed(self._add_holiday_may_5("Senior Citizens Day"))

        if self._year in {2011, 2012}:
            # Memorial Day.
            self._add_holiday_last_mon_of_may("Memorial Day")

        if self._year >= 2018:
            # President's Day.
            self._add_observed(self._add_holiday_jun_1("President's Day"))

        # Constitution Day.
        self._add_observed(self._add_holiday_jul_9("Constitution Day"))

        # Labor Day.
        self._add_holiday_1st_mon_of_sep("Labor Day")

        if self._year >= 2018:
            # Independence Day.
            self._add_observed(self._add_holiday_oct_1("Independence Day"))

        # United Nations Day.
        self._add_observed(self._add_united_nations_day("United Nations Day"))

        # Thanksgiving Day.
        self._add_holiday_4th_thu_of_nov("Thanksgiving Day")

        if self._year >= 2017:
            # Family Day.
            self._add_holiday_4th_fri_of_nov("Family Day")

        # Christmas Day.
        self._add_observed(self._add_christmas_day("Christmas Day"))


class PW(Palau):
    pass


class PLW(Palau):
    pass


class PalauStaticHolidays:
    """
    Sources:
     - https://www.facebook.com/photo?fbid=1774513196034105&set=a.175933635892077
     - https://www.facebook.com/photo/?fbid=1794692910682800&set=a.175933635892077
     - https://www.facebook.com/photo/?fbid=1408133829338712&set=a.175933635892077
    """

    special_armed_forces_holidays = {
        2020: (NOV, 11, "Veterans Day"),
    }

    special_half_day_holidays = {
        2019: (SEP, 30, "Preparation for the 25th Independence Day of the Republic of Palau"),
    }

    special_public_holidays = {
        2020: (NOV, 3, "National Day of Democracy"),
    }



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see AUTHORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from holidays.calendars.gregorian import JUL
from holidays.countries.united_kingdom import UnitedKingdom, UnitedKingdomStaticHolidays
from holidays.groups import ChristianHolidays, InternationalHolidays, StaticHolidays
from holidays.observed_holiday_base import ObservedHolidayBase, SAT_SUN_TO_NEXT_MON


class IsleOfMan(UnitedKingdom):
    """Using existing code in UnitedKingdom for now."""

    country = "IM"
    subdivisions = ()  # Override UnitedKingdom subdivisions.

    def __init__(self, *args, **kwargs):  # Override UnitedKingdom __init__().
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        StaticHolidays.__init__(self, UnitedKingdomStaticHolidays)
        kwargs.setdefault("observed_rule", SAT_SUN_TO_NEXT_MON)
        ObservedHolidayBase.__init__(self, *args, **kwargs)

    def _populate_public_holidays(self) -> None:
        super()._populate_public_holidays()
        # Easter Monday
        self._add_easter_monday("Easter Monday")

        # Whit Monday.
        if self._year <= 1970:
            self._add_whit_monday("Whit Monday")

        # Late Summer bank holiday (last Monday in August)
        if self._year >= 1971:
            self._add_holiday_last_mon_of_aug("Late Summer Bank Holiday")

        # Isle of Man exclusive holidays
        # TT bank holiday (first Friday in June)
        self._add_holiday_1st_fri_of_jun("TT Bank Holiday")

        # Tynwald Day
        # Move to the next Monday if falls on a weekend.
        dt = date(self._year, JUL, 5)
        self._add_holiday(
            "Tynwald Day",
            self._get_observed_date(dt, SAT_SUN_TO_NEXT_MON) if self._year >= 1992 else dt,
        )


class IM(IsleOfMan):
    pass


class IMN(IsleOfMan):
    pass



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see AUTHORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.groups import ChristianHolidays, InternationalHolidays
from holidays.holiday_base import HolidayBase


class Lithuania(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    Lithuania holidays.

    References:
    - https://en.wikipedia.org/wiki/Public_holidays_in_Lithuania
    - https://www.kalendorius.today/
    """

    country = "LT"
    default_language = "lt"
    supported_languages = ("en_US", "lt", "uk")

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self) -> None:
        if self._year <= 1989:
            return None

        # New Year's Day.
        self._add_new_years_day(tr("Naujųjų metų diena"))

        # Day of Restoration of the State of Lithuania.
        self._add_holiday_feb_16(tr("Lietuvos valstybės atkūrimo diena"))

        # Day of Restoration of Independence of Lithuania.
        self._add_holiday_mar_11(tr("Lietuvos nepriklausomybės atkūrimo diena"))

        # Easter Sunday.
        self._add_easter_sunday(tr("Šv. Velykos"))

        # Easter Monday.
        self._add_easter_monday(tr("Antroji šv. Velykų diena"))

        # International Workers' Day.
        self._add_labor_day(tr("Tarptautinė darbo diena"))

        # Mother's Day.
        self._add_holiday_1st_sun_of_may(tr("Motinos diena"))

        # Father's Day.
        self._add_holiday_1st_sun_of_jun(tr("Tėvo diena"))

        if self._year >= 2003:
            # Day of Dew and Saint John.
            self._add_saint_johns_day(tr("Rasos ir Joninių diena"))

        if self._year >= 1991:
            self._add_holiday_jul_6(
                # Statehood Day.
                tr(
                    "Valstybės (Lietuvos karaliaus Mindaugo karūnavimo) "
                    "ir Tautiškos giesmės diena"
                ),
            )

        # Assumption Day.
        self._add_assumption_of_mary_day(tr("Žolinė (Švč. Mergelės Marijos ėmimo į dangų diena)"))

        # All Saints' Day.
        self._add_all_saints_day(tr("Visų Šventųjų diena"))

        if self._year >= 2020:
            # All Souls' Day.
            self._add_all_souls_day(tr("Mirusiųjų atminimo (Vėlinių) diena"))

        # Christmas Eve.
        self._add_christmas_eve(tr("Kūčių diena"))

        # Christmas Day.
        self._add_christmas_day(tr("Šv. Kalėdų pirma diena"))

        # Second Day of Christmas.
        self._add_christmas_day_two(tr("Šv. Kalėdų antra diena"))


class LT(Lithuania):
    pass


class LTU(Lithuania):
    pass



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see AUTHORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars import _HebrewLunisolar
from holidays.calendars.hebrew import (
    HANUKKAH,
    INDEPENDENCE_DAY,
    LAG_BAOMER,
    PASSOVER,
    PURIM,
    ROSH_HASHANAH,
    SHAVUOT,
    SUKKOT,
    YOM_KIPPUR,
)
from holidays.constants import OPTIONAL, PUBLIC, SCHOOL
from holidays.observed_holiday_base import (
    ObservedHolidayBase,
    MON_TO_NEXT_TUE,
    THU_TO_PREV_WED,
    FRI_TO_PREV_WED,
    FRI_TO_PREV_THU,
    SAT_TO_PREV_THU,
    SAT_TO_NEXT_SUN,
    SUN_TO_NEXT_MON,
)


class Israel(ObservedHolidayBase):
    """
    Israel holidays.

    References:
      - https://en.wikipedia.org/wiki/Public_holidays_in_Israel
      - https://web.archive.org/web/20190923042619/https://www.knesset.gov.il/laws/special/heb/jerusalem_day_law.htm  # noqa: E501

    """

    country = "IL"
    default_language = "he"
    # %s (observed).
    observed_label = tr("(נצפה) %s")
    supported_categories = (OPTIONAL, PUBLIC, SCHOOL)
    supported_languages = ("en_US", "he", "uk")

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("observed_rule", FRI_TO_PREV_THU + SAT_TO_PREV_THU)
        super().__init__(*args, **kwargs)

    def _get_holiday(self, holiday: str):
        return _HebrewLunisolar._get_holiday(holiday, self._year)

    def _populate_public_holidays(self):
        if self._year <= 1947:
            return None

        if self._year >= 2101:
            raise NotImplementedError

        # Rosh Hashanah (New Year).
        name = tr("ראש השנה")
        rosh_hashanah_dt = self._get_holiday(ROSH_HASHANAH)
        self._add_holiday(name, rosh_hashanah_dt)
        self._add_holiday(name, rosh_hashanah_dt + td(days=+1))

        # Yom Kippur (Day of Atonement).
        self._add_holiday(tr("יום כיפור"), self._get_holiday(YOM_KIPPUR))

        # Sukkot (Feast of Tabernacles).
        name = tr("סוכות")
        sukkot_dt = self._get_holiday(SUKKOT)
        self._add_holiday(name, sukkot_dt)
        # Simchat Torah / Shemini Atzeret.
        self._add_holiday(tr("שמחת תורה/שמיני עצרת"), sukkot_dt + td(days=+7))

        passover_dt = self._get_holiday(PASSOVER)
        # Pesach (Passover).
        self._add_holiday(tr("פסח"), passover_dt)
        # Shvi'i shel Pesach (Seventh day of Passover)
        self._add_holiday(tr("שביעי של פסח"), passover_dt + td(days=+6))

        # Yom Ha-Atzmaut (Independence Day).
        name = tr("יום העצמאות")
        independence_day_dt = self._get_holiday(INDEPENDENCE_DAY)
        rule = FRI_TO_PREV_THU + SAT_TO_PREV_THU
        if self._year >= 2004:
            rule += MON_TO_NEXT_TUE
        is_observed, _ = self._add_observed(independence_day_dt, name, rule)
        if not is_observed:
            self._add_holiday(name, independence_day_dt)

        # Shavuot.
        self._add_holiday(tr("שבועות"), self._get_holiday(SHAVUOT))

    def _populate_optional_holidays(self):
        if self._year <= 1947:
            return None

        if self._year >= 2101:
            raise NotImplementedError

        sukkot_dt = self._get_holiday(SUKKOT)
        for offset in range(1, 6):
            # Chol HaMoed Sukkot (Feast of Tabernacles holiday).
            self._add_holiday(tr("חול המועד סוכות"), sukkot_dt + td(days=offset))

        if self._year >= 2008:
            # Sigd.
            self._add_holiday(tr("סיגד"), self._get_holiday(YOM_KIPPUR) + td(days=+49))

        # Purim.
        self._add_holiday(tr("פורים"), self._get_holiday(PURIM))

        passover_dt = self._get_holiday(PASSOVER)
        for offset in range(1, 6):
            # Chol HaMoed Pesach (Passover holiday).
            self._add_holiday(tr("חול המועד פסח"), passover_dt + td(days=offset))

        if self._year >= 1963:
            # Yom Hazikaron (Fallen Soldiers and Victims of Terrorism Remembrance Day).
            name = tr("יום הזיכרון לחללי מערכות ישראל ונפגעי פעולות האיבה")
            remembrance_day_dt = self._get_holiday(INDEPENDENCE_DAY) + td(days=-1)
            rule = THU_TO_PREV_WED + FRI_TO_PREV_WED
            if self._year >= 2004:
                rule += SUN_TO_NEXT_MON
            is_observed, _ = self._add_observed(remembrance_day_dt, name, rule)
            if not is_observed:
                self._add_holiday(name, remembrance_day_dt)

        if self._year >= 1998:
            # Yom Yerushalayim (Jerusalem Day).
            self._add_holiday(tr("יום ירושלים"), self._get_holiday(LAG_BAOMER) + td(days=+10))

        # Tisha B'Av (Tisha B'Av, fast).
        name = tr("תשעה באב")
        tisha_bav_dt = self._get_holiday("TISHA_BAV")
        is_observed, _ = self._add_observed(tisha_bav_dt, name, SAT_TO_NEXT_SUN)
        if not is_observed:
            self._add_holiday(name, tisha_bav_dt)

    def _populate_school_holidays(self):
        if self._year <= 1947:
            return None

        if self._year >= 2101:
            raise NotImplementedError

        sukkot_dt = self._get_holiday(SUKKOT)
        for offset in range(1, 6):
            # Chol HaMoed Sukkot (Feast of Tabernacles holiday).
            self._add_holiday(tr("חול המועד סוכות"), sukkot_dt + td(days=offset))

        for year in (self._year - 1, self._year):
            hanukkah_dt = _HebrewLunisolar._get_holiday(HANUKKAH, year)
            for offset in range(8):
                # Hanukkah.
                self._add_holiday(tr("חנוכה"), hanukkah_dt + td(days=offset))

        # Ta`anit Ester (Fast of Esther).
        name = tr("תענית אסתר")
        purim_dt = self._get_holiday(PURIM)
        taanit_ester_dt = purim_dt + td(days=-1)
        is_observed, _ = self._add_observed(taanit_ester_dt, name, SAT_TO_PREV_THU)
        if not is_observed:
            self._add_holiday(name, taanit_ester_dt)

        # Purim.
        self._add_holiday(tr("פורים"), purim_dt)

        passover_dt = self._get_holiday(PASSOVER)
        for offset in range(1, 6):
            # Chol HaMoed Pesach (Passover holiday).
            self._add_holiday(tr("חול המועד פסח"), passover_dt + td(days=offset))

        # Lag Ba'omer (Lag BaOmer).
        self._add_holiday(tr('ל"ג בעומר'), self._get_holiday(LAG_BAOMER))


class IL(Israel):
    pass


class ISR(Israel):
    pass



#  holidays
#  --------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: Vacanza Team and individual contributors (see AUTHORS file)
#           dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/vacanza/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars.gregorian import GREGORIAN_CALENDAR, JAN, MAR, APR, MAY, JUN, JUL, NOV, DEC
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.groups import ChristianHolidays, InternationalHolidays, StaticHolidays
from holidays.holiday_base import HolidayBase


class Belarus(HolidayBase, ChristianHolidays, InternationalHolidays, StaticHolidays):
    """
    Belarus holidays.

    References:
        - http://president.gov.by/en/holidays_en/
        - http://www.belarus.by/en/about-belarus/national-holidays
        - http://laws.newsby.org/documents/ukazp/pos05/ukaz05806.htm
        - http://president.gov.by/uploads/documents/2019/464uk.pdf
        - https://ru.wikipedia.org/wiki/%D0%9F%D1%80%D0%B0%D0%B7%D0%B4%D0%BD%D0%B8%D0%BA%D0%B8_%D0%91%D0%B5%D0%BB%D0%BE%D1%80%D1%83%D1%81%D1%81%D0%B8%D0%B8  # noqa: E501
    """

    country = "BY"
    default_language = "be"
    supported_languages = ("be", "en_US")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        StaticHolidays.__init__(self, BelarusStaticHolidays)
        super().__init__(*args, **kwargs)

    def _populate_public_holidays(self):
        # The current set of holidays actual from 1998.
        if self._year <= 1997:
            return None

        # New Year's Day.
        self._add_new_years_day(tr("Новы год"))

        if self._year >= 2020:
            self._add_new_years_day_two(tr("Новы год"))

        # Orthodox Christmas Day.
        self._add_christmas_day(tr("Нараджэнне Хрыстова (праваслаўнае Раство)"))

        # Women's Day.
        self._add_womens_day(tr("Дзень жанчын"))

        # Radunitsa (Day of Rejoicing).
        self._add_rejoicing_day(tr("Радаўніца"))

        # Labor Day.
        self._add_labor_day(tr("Свята працы"))

        # Victory Day.
        self._add_world_war_two_victory_day(tr("Дзень Перамогі"))

        # Independence Day.
        self._add_holiday_jul_3(tr("Дзень Незалежнасці Рэспублікі Беларусь (Дзень Рэспублікі)"))

        # October Revolution Day.
        self._add_holiday_nov_7(tr("Дзень Кастрычніцкай рэвалюцыі"))

        # Catholic Christmas Day.
        self._add_christmas_day(tr("Нараджэнне Хрыстова (каталіцкае Раство)"), GREGORIAN_CALENDAR)


class BY(Belarus):
    pass


class BLR(Belarus):
    pass


class BelarusStaticHolidays:
    # Date format (see strftime() Format Codes)
    substituted_date_format = tr("%d.%m.%Y")
    # Day off (substituted from %s).
    substituted_label = tr("Выходны (перанесены з %s)")
    special_public_holidays = {
        1998: (
            (JAN, 2, JAN, 10),
            (APR, 27, APR, 25),
        ),
        1999: (
            (JAN, 8, JAN, 16),
            (APR, 19, APR, 17),
        ),
        2000: (
            (MAY, 8, MAY, 13),
            (NOV, 6, NOV, 11),
        ),
        2001: (
            (JAN, 2, JAN, 20),
            (MAR, 9, MAR, 3),
            (APR, 23, APR, 21),
            (APR, 30, APR, 28),
            (JUL, 2, JUL, 7),
            (DEC, 24, DEC, 22),
            (DEC, 31, DEC, 29),
        ),
        2002: (
            (JAN, 2, JAN, 5),
            (MAY, 10, MAY, 18),
            (NOV, 8, NOV, 16),
        ),
        2003: (
            (JAN, 6, JAN, 4),
            (MAY, 5, MAY, 3),
        ),
        2004: (
            (JAN, 2, JAN, 10),
            (JAN, 5, JAN, 17),
            (JAN, 6, JAN, 31),
            (APR, 19, APR, 17),
        ),
        2005: (MAR, 7, MAR, 12),
        2006: (
            (JAN, 2, JAN, 21),
            (MAY, 8, MAY, 6),
            (NOV, 6, NOV, 4),
        ),
        2007: (
            (JAN, 2, DEC, 30, 2006),
            (MAR, 9, MAR, 17),
            (APR, 16, APR, 14),
            (APR, 30, MAY, 5),
            (JUL, 2, JUL, 7),
            (DEC, 24, DEC, 22),
            (DEC, 31, DEC, 29),
        ),
        2008: (
            (JAN, 2, JAN, 12),
            (MAY, 5, MAY, 3),
            (JUL, 4, JUN, 28),
            (DEC, 26, DEC, 20),
        ),
        2009: (
            (JAN, 2, JAN, 10),
            (APR, 27, APR, 25),
        ),
        2010: (
            (JAN, 8, JAN, 23),
            (APR, 12, APR, 17),
            (MAY, 10, MAY, 15),
        ),
        2011: (
            (MAR, 7, MAR, 12),
            (MAY, 2, MAY, 14),
        ),
        2012: (
            (MAR, 9, MAR, 11),
            (APR, 23, APR, 28),
            (JUL, 2, JUN, 30),
            (DEC, 24, DEC, 22),
            (DEC, 31, DEC, 29),
        ),
        2013: (
            (JAN, 2, JAN, 5),
            (MAY, 10, MAY, 18),
        ),
        2014: (
            (JAN, 2, JAN, 4),
            (JAN, 6, JAN, 11),
            (APR, 30, MAY, 3),
            (JUL, 4, JUL, 12),
            (DEC, 26, DEC, 20),
        ),
        2015: (
            (JAN, 2, JAN, 10),
            (APR, 20, APR, 25),
        ),
        2016: (
            (JAN, 8, JAN, 16),
            (MAR, 7, MAR, 5),
        ),
        2017: (
            (JAN, 2, JAN, 21),
            (APR, 24, APR, 29),
            (MAY, 8, MAY, 6),
            (NOV, 6, NOV, 4),
        ),
        2018: (
            (JAN, 2, JAN, 20),
            (MAR, 9, MAR, 3),
            (APR, 16, APR, 14),
            (APR, 30, APR, 28),
            (JUL, 2, JUL, 7),
            (DEC, 24, DEC, 22),
            (DEC, 31, DEC, 29),
        ),
        2019: (
            (MAY, 6, MAY, 4),
            (MAY, 8, MAY, 11),
            (NOV, 8, NOV, 16),
        ),
        2020: (
            (JAN, 6, JAN, 4),
            (APR, 27, APR, 4),
        ),
        2021: (
            (JAN, 8, JAN, 16),
            (MAY, 10, MAY, 15),
        ),
        2022: (
            (MAR, 7, MAR, 12),
            (MAY, 2, MAY, 14),
        ),
        2023: (
            (APR, 24, APR, 29),
            (MAY, 8, MAY, 13),
            (NOV, 6, NOV, 11),
        ),
    }
